"""Interactive map of one or more GeoDataFrames with layers that can be toggles on/off.

This module holds the Explore class, which is the basis for the explore, samplemap and
clipmap functions from the 'maps' module.
"""
import os
import warnings
from collections.abc import Iterable
from numbers import Number
from statistics import mean

import branca as bc
import folium
import matplotlib
import numpy as np
import pandas as pd
import xyzservices
from folium import plugins
from geopandas import GeoDataFrame
from IPython.display import display
from jinja2 import Template
from pandas.api.types import is_datetime64_any_dtype
from shapely import Geometry
from shapely.geometry import LineString

from ..geopandas_tools.conversion import from_4326, to_gdf
from ..geopandas_tools.general import clean_geoms, make_all_singlepart
from ..geopandas_tools.geometry_types import get_geom_type, to_single_geom_type
from .httpserver import run_html_server
from .map import Map
from .tilesources import kartverket, xyz


# the geopandas._explore raises a deprication warning. Ignoring for now.
warnings.filterwarnings(
    action="ignore", category=matplotlib.MatplotlibDeprecationWarning
)
pd.options.mode.chained_assignment = None


# gray for NaNs
NAN_COLOR = "#969696"


# cols to not show when hovering over geometries (tooltip)
COLS_TO_DROP = ["color", "col_as_int", "geometry"]


# from geopandas
_MAP_KWARGS = [
    "location",
    "prefer_canvas",
    "no_touch",
    "disable_3d",
    "png_enabled",
    "zoom_control",
    "crs",
    "zoom_start",
    "left",
    "top",
    "position",
    "min_zoom",
    "max_zoom",
    "min_lat",
    "max_lat",
    "min_lon",
    "max_lon",
    "max_bounds",
]


class MeasureControlFix(plugins.MeasureControl):
    """Monkey-patch to fix a bug in the lenght measurement control.

    Kudos to abewartech (https://github.com/ljagis/leaflet-measure/issues/171).
    """

    _template = Template(
        """
{% macro script(this, kwargs) %}
    L.Control.Measure.include({ _setCaptureMarkerIcon: function () { this._captureMarker.options.autoPanOnFocus = false; this._captureMarker.setIcon( L.divIcon({ iconSize: this._map.getSize().multiplyBy(2), }), ); }, });
    var {{ this.get_name() }} = new L.Control.Measure(
        {{ this.options|tojson }});
    {{this._parent.get_name()}}.addControl({{this.get_name()}});

{% endmacro %}
    """
    )

    def __init__(self, active_color="red", completed_color="red", **kwargs):
        super().__init__(
            active_color=active_color, completed_color=completed_color, **kwargs
        )


def to_tile(tile: str | xyzservices.TileProvider, max_zoom: int) -> folium.TileLayer:
    common_bgmaps = {
        "openstreetmap": folium.TileLayer(
            "OpenStreetMap", min_zoom=0, max_zoom=max_zoom
        ),
        "grunnkart": kartverket.norges_grunnkart,
        "gråtone": kartverket.norges_grunnkart_gråtone,
        "norge_i_bilder": kartverket.norge_i_bilder,
        "dark": xyz.CartoDB.DarkMatter,
        "voyager": xyz.CartoDB.Voyager,
        "strava": xyz.Strava.All,
    }
    try:
        name = tile["name"]
    except TypeError:
        name = tile

    if not isinstance(tile, str):
        try:
            return folium.TileLayer(tile, name=name, max_zoom=max_zoom)
        except TypeError:
            return folium.TileLayer(tile, max_zoom=max_zoom)

    try:
        provider = common_bgmaps[tile.lower()]
    except KeyError:
        provider = xyzservices.providers.query_name(tile)

    if isinstance(provider, folium.TileLayer):
        return provider

    if isinstance(provider, xyzservices.TileProvider):
        attr = provider.html_attribution
        provider = provider.build_url(scale_factor="{r}")
    else:
        try:
            attr = provider["attr"]
        except (AttributeError, TypeError):
            attr = None

    return folium.TileLayer(provider, name=name, attr=attr, max_zoom=max_zoom)


class Explore(Map):
    # class attribute that can be overridden locally
    tiles = ("OpenStreetMap", "dark", "norge_i_bilder", "grunnkart")

    def __init__(
        self,
        *gdfs,
        mask=None,
        column: str | None = None,
        popup: bool = True,
        max_zoom: int = 30,
        smooth_factor: float = 1.5,
        browser: bool = False,
        prefer_canvas: bool = True,
        measure_control: bool = True,
        geocoder: bool = True,
        save=None,
        show: bool | Iterable[bool] | None = None,
        text: str | None = None,
        **kwargs,
    ):
        self.popup = popup
        self.max_zoom = max_zoom
        self.smooth_factor = smooth_factor
        self.prefer_canvas = prefer_canvas
        self.measure_control = measure_control
        self.geocoder = geocoder
        self.save = save
        self.mask = mask
        self.text = text

        self.browser = browser
        if not self.browser and "show_in_browser" in kwargs:
            self.browser = kwargs.pop("show_in_browser")
        if not self.browser and "in_browser" in kwargs:
            self.browser = kwargs.pop("in_browser")

        if show is None:
            show_was_none = True
            show = True
        else:
            show_was_none = False

        super().__init__(*gdfs, column=column, show=show, **kwargs)

        if self.gdfs is None:
            return

        # stringify or remove columns not renerable by leaflet (list, geometry etc.)
        new_gdfs, show_new = [], []
        for gdf, show in zip(self.gdfs, self.show, strict=True):
            try:
                gdf = gdf.reset_index()
            except Exception:
                pass
            for col in gdf.columns:
                if is_datetime64_any_dtype(gdf[col]):
                    try:
                        gdf[col] = [str(x) for x in gdf[col].dt.round("d")]
                    except Exception:
                        gdf = gdf.drop(col, axis=1)
                    continue

                if not len(gdf.loc[gdf[col].notna()]):
                    continue
                if not isinstance(
                    gdf.loc[gdf[col].notna(), col].iloc[0], (Number, str, Geometry)
                ) or (
                    col != gdf._geometry_column_name
                    and isinstance(gdf.loc[gdf[col].notna(), col].iloc[0], (Geometry))
                ):
                    try:
                        gdf[col] = gdf[col].astype(str).fillna(pd.NA)
                    except Exception:
                        gdf = gdf.drop(col, axis=1)

            try:
                gdf.index = gdf.index.astype(str)
            except Exception:
                pass
            new_gdfs.append(gdf)
            show_new.append(show)
        self._gdfs = new_gdfs
        self._gdf = pd.concat(new_gdfs, ignore_index=True)
        self.show = show_new

        if show_was_none and len(self._gdfs) > 6:
            self.show = [False] * len(self._gdfs)

        if self._is_categorical:
            if len(self.gdfs) == 1:
                self._split_categories()
        else:
            if not self._cmap:
                self._cmap = "viridis"
            self.cmap_start = kwargs.pop("cmap_start", 0)
            self.cmap_stop = kwargs.pop("cmap_stop", 256)

        if self._gdf.crs is None:
            self.kwargs["crs"] = "Simple"

        self.original_crs = self.gdf.crs

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def explore(
        self, column: str | None = None, center=None, size=None, **kwargs
    ) -> None:
        if not any(len(gdf) for gdf in self._gdfs):
            warnings.warn("None of the GeoDataFrames have rows.")
            return
        if column:
            self._column = column
            self._update_column()
            kwargs.pop("column", None)

        if self.mask is not None:
            return self.clipmap(mask=self.mask, column=self._column, **kwargs)

        if center is None:
            self.to_show = self._gdfs
            self._explore(**kwargs)
            return

        size = size if size else 1000

        centerpoint = (
            to_gdf(center, crs=self.crs)
            if not isinstance(center, GeoDataFrame)
            else center
        )

        gdfs: tuple[GeoDataFrame] = ()
        for gdf in self._gdfs:
            keep_geom_type = False if get_geom_type(gdf) == "mixed" else True
            gdf = gdf.clip(centerpoint.buffer(size), keep_geom_type=keep_geom_type)
            gdfs = gdfs + (gdf,)
        self._gdfs = gdfs
        self._gdf = pd.concat(gdfs, ignore_index=True)

        self._get_unique_values()

        self._explore(**kwargs)

    def samplemap(
        self,
        size: int = 1000,
        column: str | None = None,
        sample_from_first: bool = True,
        **kwargs,
    ) -> None:
        if column:
            self._column = column
            self._update_column()
            kwargs.pop("column", None)

        if sample_from_first:
            sample = self._gdfs[0].sample(1)
        else:
            sample = self._gdf.sample(1)

        # convert lines to polygons
        if get_geom_type(sample) == "line":
            sample["geometry"] = sample.buffer(1)

        if get_geom_type(sample) == "polygon":
            random_point = sample.sample_points(size=1)

        # if point or mixed geometries
        else:
            random_point = sample.centroid

        self.center = (random_point.geometry.iloc[0].x, random_point.geometry.iloc[0].y)
        print(f"center={self.center}, size={size}")

        gdfs: tuple[GeoDataFrame] = ()
        for gdf in self._gdfs:
            gdf = gdf.clip(random_point.buffer(size))
            gdfs = gdfs + (gdf,)
        self._gdfs = gdfs
        self._gdf = pd.concat(gdfs, ignore_index=True)

        self._get_unique_values()
        self._explore(**kwargs)

    def clipmap(
        self,
        mask,
        column: str | None = None,
        **kwargs,
    ) -> None:
        if column:
            self._column = column
            self._update_column()
            kwargs.pop("column", None)

        gdfs: tuple[GeoDataFrame] = ()
        for gdf in self._gdfs:
            gdf = gdf.clip(mask)
            collections = gdf.loc[gdf.geom_type == "GeometryCollection"]
            if len(collections):
                collections = make_all_singlepart(collections)
                gdf = pd.concat([gdf, collections], ignore_index=False)
            gdfs = gdfs + (gdf,)
        self._gdfs = gdfs
        self._gdf = pd.concat(gdfs, ignore_index=True)
        self._explore(**kwargs)

    def _explore(self, **kwargs):
        self.kwargs = self.kwargs | kwargs

        if self._is_categorical:
            self._create_categorical_map()
        else:
            self._create_continous_map()

        if self.save:
            with open(os.getcwd() + "/" + self.save.strip(".html") + ".html", "w") as f:
                f.write(self.map._repr_html_())
        elif self.browser:
            run_html_server(self.map._repr_html_())
        else:
            display(self.map)

    def _split_categories(self):
        new_gdfs, new_labels, new_shows = [], [], []
        for cat in self._unique_values:
            gdf = self.gdf.loc[self.gdf[self.column] == cat]
            new_gdfs.append(gdf)
            new_labels.append(cat)
            new_shows.append(self.show[0])
        self._gdfs = new_gdfs
        self._gdf = pd.concat(new_gdfs, ignore_index=True)
        self.labels = new_labels
        self.show = new_shows

    def _to_single_geom_type(self, gdf) -> GeoDataFrame:
        gdf = clean_geoms(gdf)

        if get_geom_type(gdf) != "mixed":
            return gdf

        geom_types = gdf.geom_type.str.lower()
        mess = "Leaflet cannot render mixed geometry types well. "

        if geom_types.str.contains("collection").any():
            mess += "Exploding geometry collections. "
            gdf = make_all_singlepart(gdf)
            geom_types = gdf.geom_type.str.lower()

        if geom_types.str.contains("polygon").any():
            mess += "Keeping only polygons."
            gdf = to_single_geom_type(gdf, geom_type="polygon")

        elif geom_types.str.contains("lin").any():
            mess += "Keeping only lines."
            gdf = to_single_geom_type(gdf, geom_type="line")

        assert get_geom_type(gdf) != "mixed", gdf.geom_type.value_counts()

        warnings.warn(mess)

        return gdf

    def _update_column(self):
        self._is_categorical = self._check_if_categorical()
        self._fillna_if_col_is_missing()
        self._gdf = pd.concat(self._gdfs, ignore_index=True)

    def _create_categorical_map(self):
        self._get_categorical_colors()

        gdf = self._prepare_gdf_for_map(self._gdf)
        self.map = self._make_folium_map(
            bounds=gdf.total_bounds,
            max_zoom=self.max_zoom,
            popup=self.popup,
            prefer_canvas=self.prefer_canvas,
            **self.kwargs,
        )

        for gdf, label, show in zip(self._gdfs, self.labels, self.show, strict=True):
            if not len(gdf):
                continue

            f = folium.FeatureGroup(name=label)

            gdf = self._to_single_geom_type(gdf)
            gdf = self._prepare_gdf_for_map(gdf)

            gjs = self._make_geojson(
                gdf,
                show=show,
                color=gdf["color"],
                tooltip=self._tooltip_cols(gdf),
                popup=self.popup,
                **{
                    key: value
                    for key, value in self.kwargs.items()
                    if key not in ["title", "tiles"]
                },
            )
            gjs.layer_name = label

            gjs.add_to(f)
            gjs.add_to(self.map)

        _categorical_legend(
            self.map,
            self._column,
            self._categories_colors_dict.keys(),
            self._categories_colors_dict.values(),
        )

        self.map.add_child(folium.LayerControl())

    def _add_tiles(
        self, mapobj: folium.Map, tiles: list[str, xyzservices.TileProvider]
    ):
        for tile in tiles:
            to_tile(tile, max_zoom=self.max_zoom).add_to(mapobj)

    def _create_continous_map(self):
        self._prepare_continous_map()
        if self.scheme:
            classified = self._classify_from_bins(self._gdf, bins=self.bins)
            classified_sequential = self._push_classification(classified)
            n_colors = len(np.unique(classified_sequential)) - any(self._nan_idx)
            unique_colors = self._get_continous_colors(n=n_colors)

        gdf = self._prepare_gdf_for_map(self._gdf)
        self.map = self._make_folium_map(
            bounds=gdf.total_bounds,
            max_zoom=self.max_zoom,
            popup=self.popup,
            prefer_canvas=self.prefer_canvas,
            **self.kwargs,
        )

        colorbar = bc.colormap.StepColormap(
            unique_colors,
            vmin=self._gdf[self._column].min(),
            vmax=self._gdf[self._column].max(),
            caption=self._column,
            index=self.bins,
        )

        for gdf, label, show in zip(self._gdfs, self.labels, self.show, strict=True):
            if not len(gdf):
                continue
            f = folium.FeatureGroup(name=label)

            gdf = self._to_single_geom_type(gdf)
            gdf = self._prepare_gdf_for_map(gdf)

            classified = self._classify_from_bins(gdf, bins=self.bins)
            try:
                colorarray = unique_colors[classified]
            except IndexError:
                classified[classified > 0] = classified[classified > 0] - 1
                colorarray = unique_colors[classified]

            gjs = self._make_geojson(
                gdf,
                show=show,
                color=colorarray,
                tooltip=self._tooltip_cols(gdf),
                popup=self.popup,
                prefer_canvas=self.prefer_canvas,
                **{
                    key: value
                    for key, value in self.kwargs.items()
                    if key not in ["title"]
                },
            )

            f.add_child(gjs)
            self.map.add_child(f)

        self.map.add_child(colorbar)
        self.map.add_child(folium.LayerControl())

    def _tooltip_cols(self, gdf: GeoDataFrame) -> list:
        if "tooltip" in self.kwargs:
            tooltip = self.kwargs.pop("tooltip")
            return tooltip
        return [col for col in gdf.columns if col not in COLS_TO_DROP]

    @staticmethod
    def _prepare_gdf_for_map(gdf):
        # convert LinearRing to LineString
        rings_mask = gdf.geom_type == "LinearRing"
        if rings_mask.any():
            gdf.geometry[rings_mask] = gdf.geometry[rings_mask].apply(
                lambda g: LineString(g)
            )

        if gdf.crs is not None and not gdf.crs.equals(4326):
            gdf = gdf.to_crs(4326)

        return gdf

    def _make_folium_map(
        self,
        bounds,
        attr=None,
        tiles=None,
        width="100%",
        height="100%",
        control_scale=True,
        map_kwds=None,
        **kwargs,
    ):
        if not map_kwds:
            map_kwds = {}

        if tiles is None:
            tiles = self.tiles

        # create folium.Map object
        # Get bounds to specify location and map extent
        location = kwargs.pop("location", None)
        if location is None:
            x = mean([bounds[0], bounds[2]])
            y = mean([bounds[1], bounds[3]])
            location = (y, x)
            if "zoom_start" in kwargs.keys():
                fit = False
            else:
                fit = True
        else:
            fit = False

        # get a subset of kwargs to be passed to folium.Map
        for i in _MAP_KWARGS:
            if i in map_kwds:
                raise ValueError(
                    f"'{i}' cannot be specified in 'map_kwds'. "
                    f"Use the '{i}={map_kwds[i]}' argument instead."
                )
        map_kwds = {
            **map_kwds,
            **{i: kwargs[i] for i in kwargs.keys() if i in _MAP_KWARGS},
        }
        map_kwds["min_zoom"] = 0
        map_kwds["max_zoom"] = kwargs.get("max_zoom", self.max_zoom)

        if isinstance(tiles, (list, tuple)):
            default_tile, *more_tiles = tiles
        else:
            default_tile, more_tiles = tiles, []

        default_tile = to_tile(default_tile, max_zoom=self.max_zoom)

        if isinstance(default_tile, xyzservices.TileProvider):
            attr = attr if attr else default_tile.html_attribution
            default_tile = default_tile.build_url(scale_factor="{r}")

        m = folium.Map(
            location=location,
            control_scale=control_scale,
            tiles=default_tile,
            attr=attr,
            width=width,
            height=height,
            **map_kwds,
        )

        self._add_tiles(m, more_tiles)

        if self.measure_control:
            MeasureControlFix(
                primary_length_unit="meters",
                secondary_length_unit="kilometers",
                primary_area_unit="sqmeters",
                secondary_area_unit="sqkilometers",
                position="bottomleft",
                capture_z_index=False,
            ).add_to(m)

        plugins.Fullscreen(
            position="topleft",
            title="Expand me",
            title_cancel="Exit me",
            force_separate_button=True,
        ).add_to(m)

        plugins.MousePosition(
            position="bottomleft",
            separator=", ",
            empty_string="NaN",
            lng_first=True,
            num_digits=8,
        ).add_to(m)

        if self.geocoder:
            plugins.Geocoder(position="topright").add_to(m)

        # fit bounds to get a proper zoom level
        if fit and "zoom_start" not in kwargs:
            m.fit_bounds([[bounds[1], bounds[0]], [bounds[3], bounds[2]]])

        if self.text:
            style = bc.element.MacroElement()
            style._template = bc.element.Template(get_textbox(self.text))
            m.get_root().add_child(style)
            # folium.LayerControl(collapsed=False).add_to(m)

        return m

    def _make_geojson(
        self,
        df,
        show: bool,
        color=None,
        tooltip=True,
        popup=False,
        highlight=True,
        marker_type=None,
        marker_kwds={},
        style_kwds={},
        highlight_kwds={},
        tooltip_kwds={},
        popup_kwds={},
        map_kwds={},
        **kwargs,
    ):
        gdf = df.copy()

        # convert LinearRing to LineString
        rings_mask = gdf.geom_type == "LinearRing"
        if rings_mask.any():
            gdf.geometry[rings_mask] = gdf.geometry[rings_mask].apply(
                lambda g: LineString(g)
            )

        if gdf.crs is None:
            kwargs["crs"] = "Simple"
        elif not gdf.crs.equals(4326):
            gdf = gdf.to_crs(4326)

        # get a subset of kwargs to be passed to folium.Map
        for i in _MAP_KWARGS:
            if i in map_kwds:
                raise ValueError(
                    f"'{i}' cannot be specified in 'map_kwds'. "
                    f"Use the '{i}={map_kwds[i]}' argument instead."
                )
        map_kwds = {
            **map_kwds,
            **{i: kwargs[i] for i in kwargs.keys() if i in _MAP_KWARGS},
        }

        for map_kwd in _MAP_KWARGS:
            kwargs.pop(map_kwd, None)

        # set default style
        if "fillOpacity" not in style_kwds:
            style_kwds["fillOpacity"] = 0.5
        if "weight" not in style_kwds:
            style_kwds["weight"] = 2
        if "style_function" in style_kwds:
            style_kwds_function = style_kwds["style_function"]
            if not callable(style_kwds_function):
                raise ValueError("'style_function' has to be a callable")
            style_kwds.pop("style_function")
        else:

            def _no_style(x):
                return {}

            style_kwds_function = _no_style

        gdf["__folium_color"] = color

        stroke_color = style_kwds.pop("color", None)
        if not stroke_color:

            def _style_column(x):
                base_style = {
                    "fillColor": x["properties"]["__folium_color"],
                    "color": x["properties"]["__folium_color"],
                    **style_kwds,
                }
                return {
                    **base_style,
                    **style_kwds_function(x),
                }

            style_function = _style_column

        if highlight:
            if "fillOpacity" not in highlight_kwds:
                highlight_kwds["fillOpacity"] = 0.75

            def _style_highlight(x):
                return {**highlight_kwds}

            highlight_function = _style_highlight
        else:
            highlight_function = None

        # define default for points
        if marker_type is None:
            marker_type = "circle_marker"

        marker = marker_type
        if isinstance(marker_type, str):
            if marker_type == "marker":
                marker = folium.Marker(**marker_kwds)
            elif marker_type == "circle":
                marker = folium.Circle(**marker_kwds)
            elif marker_type == "circle_marker":
                marker_kwds["radius"] = marker_kwds.get("radius", 2)
                marker_kwds["fill"] = marker_kwds.get("fill", True)
                marker = folium.CircleMarker(**marker_kwds)
            else:
                raise ValueError(
                    "Only 'marker', 'circle', and 'circle_marker' are "
                    "supported as marker values"
                )

        # prepare tooltip and popup
        if isinstance(gdf, GeoDataFrame):
            # add named index to the tooltip
            if gdf.index.name is not None:
                gdf = gdf.reset_index()
            # specify fields to show in the tooltip
            tooltip = _tooltip_popup("tooltip", tooltip, gdf, **tooltip_kwds)
            popup = _tooltip_popup("popup", popup, gdf, **popup_kwds)
        else:
            tooltip = None
            popup = None

        gdf_as_json = gdf.__geo_interface__

        return folium.GeoJson(
            gdf_as_json,
            tooltip=tooltip,
            popup=popup,
            marker=marker,
            style_function=style_function,
            highlight_function=highlight_function,
            smooth_factor=self.smooth_factor,
            show=show,
            **kwargs,
        )


def _tooltip_popup(type, fields, gdf, **kwds):
    """get tooltip or popup"""

    # specify fields to show in the tooltip
    if fields is False or fields is None or fields == 0:
        return None
    else:
        if fields is True:
            fields = gdf.columns.drop(gdf.geometry.name).to_list()
        elif isinstance(fields, int):
            fields = gdf.columns.drop(gdf.geometry.name).to_list()[:fields]
        elif isinstance(fields, str):
            fields = [fields]

    for field in ["__plottable_column", "__folium_color"]:
        if field in fields:
            fields.remove(field)

    # Cast fields to str
    fields = list(map(str, fields))
    if type == "tooltip":
        return folium.GeoJsonTooltip(fields, **kwds)
    elif type == "popup":
        return folium.GeoJsonPopup(fields, **kwds)


def _categorical_legend(m, title, categories, colors):
    """
    Add categorical legend to a map

    The implementation is using the code originally written by Michel Metran
    (@michelmetran) and released on GitHub
    (https://github.com/michelmetran/package_folium) under MIT license.

    Copyright (c) 2020 Michel Metran

    Parameters
    ----------
    m : folium.Map
        Existing map instance on which to draw the plot
    title : str
        title of the legend (e.g. column name)
    categories : list-like
        list of categories
    colors : list-like
        list of colors (in the same order as categories)
    """

    # Header to Add
    head = """
    {% macro header(this, kwargs) %}
    <script src="https://code.jquery.com/ui/1.12.1/jquery-ui.js"></script>
    <script>$( function() {
        $( ".maplegend" ).draggable({
            start: function (event, ui) {
                $(this).css({
                    right: "auto",
                    top: "auto",
                    bottom: "auto"
                });
            }
        });
    });
    </script>
    <style type='text/css'>
      .maplegend {
        position: absolute;
        z-index:9999;
        background-color: rgba(255, 255, 255, .8);
        border-radius: 5px;
        box-shadow: 0 0 15px rgba(0,0,0,0.2);
        padding: 10px;
        font: 12px/14px Arial, Helvetica, sans-serif;
        right: 10px;
        bottom: 20px;
      }
      .maplegend .legend-title {
        text-align: left;
        margin-bottom: 5px;
        font-weight: bold;
        }
      .maplegend .legend-scale ul {
        margin: 0;
        margin-bottom: 0px;
        padding: 0;
        float: left;
        list-style: none;
        }
      .maplegend .legend-scale ul li {
        list-style: none;
        margin-left: 0;
        line-height: 16px;
        margin-bottom: 2px;
        }
      .maplegend ul.legend-labels li span {
        display: block;
        float: left;
        height: 14px;
        width: 14px;
        margin-right: 5px;
        margin-left: 0;
        border: 0px solid #ccc;
        }
      .maplegend .legend-source {
        color: #777;
        clear: both;
        }
      .maplegend a {
        color: #777;
        }
    </style>
    {% endmacro %}
    """
    import branca as bc

    # Add CSS (on Header)
    macro = bc.element.MacroElement()
    macro._template = bc.element.Template(head)
    m.get_root().add_child(macro)

    body = f"""
    <div id='maplegend {title}' class='maplegend'>
        <div class='legend-title'>{title}</div>
        <div class='legend-scale'>
            <ul class='legend-labels'>"""

    # Loop Categories
    for label, color in zip(categories, colors, strict=True):
        body += f"""
                <li><span style='background:{color}'></span>{label}</li>"""

    body += """
            </ul>
        </div>
    </div>
    """

    # Add Body
    body = bc.element.Element(body, "legend")
    m.get_root().html.add_child(body)


def get_textbox(text: str) -> str:
    return f"""
{{% macro html(this, kwargs) %}}
<!doctype html>
<html lang="en">
  <head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>Textbox Project</title>
    <link rel="stylesheet" href="//code.jquery.com/ui/1.12.1/themes/base/jquery-ui.css">
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.2.1/css/all.min.css" integrity="sha512-MV7K8+y+gLIBoVD59lQIYicR65iaqukzvf/nwasF0nqhPay5w/9lJmVM2hMDcnK1OnMGCdVK+iQrJ7lzPJQd1w==" crossorigin="anonymous" referrerpolicy="no-referrer"/>
    <script src="https://code.jquery.com/jquery-1.12.4.js"></script>
    <script src="https://code.jquery.com/ui/1.12.1/jquery-ui.js"></script>

    <script>
      $( function() {{
        $( "#textbox" ).draggable({{
          start: function (event, ui) {{
            $(this).css({{
              right: "auto",
              top: "auto",
              bottom: "auto"
            }});
          }}
        }});
      }});
    </script>
  </head>

  <body>
    <div id="textbox" class="textbox">
      <div class="textbox-content">
        <p>{text}</p>
      </div>
    </div>

</body>
</html>

<style type='text/css'>
  .textbox {{
    position: absolute;
    z-index:9999;
    border-radius:4px;
    background: white;
    padding: 1px;
    font-size:11px;
    left: 20px;
    top: 50%;
    color: black;
  }}
</style>
{{% endmacro %}}
"""



"""Network class with methods for manipulating line geometries.

The module includes functions for cutting and splitting lines, filling holes in the
network, finding and removing isolated network islands and creating unique node ids.
"""

import warnings
from copy import copy, deepcopy

import numpy as np
from geopandas import GeoDataFrame
from pandas import DataFrame
from shapely import line_merge

from ..exceptions import ZeroLinesError
from ..geopandas_tools.general import clean_geoms
from ..geopandas_tools.geometry_types import to_single_geom_type
from .nodes import make_node_ids


class Network:
    """Class used in NetworkAnalysis."""

    def __init__(self, gdf: GeoDataFrame):
        """The lines are fixed, welded together rowwise and exploded. Creates node-ids.

        Raises:
            TypeError: If 'gdf' is not of type GeoDataFrame.
            ZeroLinesError: If 'gdf' has zero rows.
        """
        if not isinstance(gdf, GeoDataFrame):
            raise TypeError(f"'lines' should be GeoDataFrame, got {type(gdf)}")

        if not len(gdf):
            raise ZeroLinesError

        self.gdf = self._prepare_network(gdf)

        self._make_node_ids()

        self._percent_bidirectional = self._check_percent_bidirectional()

    def _make_node_ids(self) -> None:
        """Gives the lines node ids and return lines (edges) and nodes.

        Takes the first and last point of each line and creates a GeoDataFrame of
        nodes (points) with a column 'node_id'. The node ids are then assigned to the
        input GeoDataFrame of lines as the columns 'source' and 'target'.

        Note:
            The lines must be singlepart linestrings.
        """
        self.gdf, self._nodes = make_node_ids(self.gdf)

    @staticmethod
    def _prepare_network(gdf: GeoDataFrame) -> GeoDataFrame:
        """Make sure there are only singlepart LineStrings in the network.

        This is needed when making node-ids based on the lines' endpoints, because
        MultiLineStrings have more than two endpoints, and LinearRings have zero.
        Rename geometry column to 'geometry',

        Args:
            gdf: GeoDataFrame with (multi)line geometries. MultiLineStrings will be
                merged, then exploded if a merge was not possible.

        Returns:
            A GeoDataFrame of line geometries.

        Raises:
            ZeroLinesError: If the GeoDataFrame has 0 rows.
        """
        gdf["idx_orig"] = gdf.index

        if gdf._geometry_column_name != "geometry":
            gdf = gdf.rename_geometry("geometry")

        gdf = clean_geoms(gdf)
        gdf = to_single_geom_type(gdf, geom_type="lines")

        if not len(gdf):
            raise ZeroLinesError

        gdf.geometry = line_merge(gdf.geometry)

        rows_now = len(gdf)
        gdf = gdf.loc[gdf.geom_type != "LinearRing"]

        if diff := rows_now - len(gdf):
            if diff == 1:
                print(f"{diff} LinearRing was removed from the network.")
            else:
                print(f"{diff} LinearRings were removed from the network.")

        rows_now = len(gdf)
        gdf = gdf.explode(ignore_index=True)

        if diff := len(gdf) - rows_now:
            if diff == 1:
                print(
                    "1 multi-geometry was split into single part geometries. "
                    "Minute column(s) will be wrong for these rows."
                )
            else:
                print(
                    f"{diff} multi-geometries were split into single part geometries. "
                    "Minute column(s) will be wrong for these rows."
                )

        return gdf

    def _check_percent_bidirectional(self) -> int:
        """Road data often have to be duplicated and flipped to make it directed.

        Here we check how.
        """
        self.gdf["meters"] = self.gdf.length.round(10).astype(str)
        no_dups = DataFrame(
            np.sort(self.gdf[["source", "target", "meters"]].values, axis=1),
            columns=[["source", "target", "meters"]],
        ).drop_duplicates()

        # back to numeric
        self.gdf["meters"] = self.gdf.length

        percent_bidirectional = len(self.gdf) / len(no_dups) * 100 - 100

        return int(round(percent_bidirectional, 0))

    def _nodes_are_up_to_date(self) -> bool:
        """Check if nodes need to be updated.

        Returns False if there are any source or target values not in the node-ids,
        or any superfluous node-ids (meaning rows have been removed from the lines
        gdf).
        """
        new_or_missing = (~self.gdf.source.isin(self._nodes.node_id)) | (
            ~self.gdf.target.isin(self._nodes.node_id)
        )

        if any(new_or_missing):
            return False

        removed = ~(
            (self._nodes.node_id.isin(self.gdf.source))
            | (self._nodes.node_id.isin(self.gdf.target))
        )

        if any(removed):
            return False

        return True

    def get_edges(self) -> list[tuple[str, str]]:
        return [
            (str(source), str(target))
            for source, target in zip(
                self.gdf["source"], self.gdf["target"], strict=True
            )
        ]

    @staticmethod
    def _create_edge_ids(
        edges: list[tuple[str, str]], weights: list[float]
    ) -> list[str]:
        """Edge identifiers represented with source and target ids and the weight."""
        return [f"{s}_{t}_{w}" for (s, t), w in zip(edges, weights, strict=True)]

    def _update_nodes_if(self):
        if not self._nodes_are_up_to_date():
            self._make_node_ids()

    @property
    def nodes(self):
        """GeoDataFrame with the network nodes (line endpoints).

        Upon instantiation of the class, a GeoDataFrame of points is created from the
        unique endpoints of the lines. The node ids are then used to make the 'source'
        and 'target' columns of the line gdf. The nodes are remade every time the
        geometries of the line gdf changes, since the 'source' and 'target' columns
        must be up to date with the actual line geometries when making the graph.
        """
        return self._nodes

    def _warn_if_undirected(self):
        """Road data often have to be duplicated and flipped to make it directed."""
        if self.percent_bidirectional > 5:
            return

        mess = (
            "Your network is likely not directed. "
            f"Only {self.percent_bidirectional:.1f} percent of the lines go both ways."
        )
        if "oneway" in [col.lower() for col in self.gdf.columns]:
            mess = mess + (
                " Try setting direction_col='oneway' in the 'make_directed_network' "
                "method"
            )
        else:
            mess = mess + "Try running 'make_directed_network'"

        warnings.warn(mess, stacklevel=2)

    @property
    def percent_bidirectional(self):
        """The percentage of lines that appear in both directions."""
        return self._percent_bidirectional

    def copy(self):
        """Returns a shallow copy of the class instance."""
        return copy(self)

    def deepcopy(self):
        """Returns a deep copy of the class instance."""
        return deepcopy(self)

    def __repr__(self) -> str:
        """The print representation."""
        cl = self.__class__.__name__
        km = int(sum(self.gdf.length) / 1000)
        return f"{cl}({km} km, percent_bidirectional={self._percent_bidirectional})"

    def __iter__(self):
        """So the attributes can be iterated through."""
        return iter(self.__dict__.items())

    def __len__(self):
        return len(self.gdf)



from contextlib import contextmanager


try:
    import dapla as dp
except ImportError:
    pass

from ._is_dapla import is_dapla


@contextmanager
def opener(path, mode="rb"):
    """Yields a gcs buffer if in Dapla, otherwise yields the path.

    Example
    -------
    >>> with opener(path) as file:
    >>>     with rasterio.open(file) as src:
    >>>         array = src.read()
    """
    if is_dapla():
        fs = dp.FileClient.get_gcs_file_system()
        yield fs.open(str(path), mode=mode)
    else:
        yield str(path)



import numpy as np

from .raster import Raster


class ElevationRaster(Raster):
    """For raster calculation on elevation images."""

    def __init__(self, raster=None, **kwargs):
        kwargs = {"band_index": 1} | kwargs
        super().__init__(raster, **kwargs)

    def gradient(self, degrees: bool = False, copy: bool = False):
        """Get the slope of an elevation raster.

        Calculates the absolute slope between the grid cells
        based on the image resolution.

        For multiband images, the calculation is done for each band.

        Args:
            degrees: If False (default), the returned values will be in ratios,
                where a value of 1 means 1 meter up per 1 meter forward. If True,
                the values will be in degrees from 0 to 90.
            copy: Whether to copy or overwrite the original Raster.
                Defaults to False to save memory.

        Returns:
            The class instance with new array values, or a copy if copy is True.

        Examples
        --------
        Making an array where the gradient to the center is always 10.

        >>> import sgis as sg
        >>> import numpy as np
        >>> arr = np.array(
        ...         [
        ...             [100, 100, 100, 100, 100],
        ...             [100, 110, 110, 110, 100],
        ...             [100, 110, 120, 110, 100],
        ...             [100, 110, 110, 110, 100],
        ...             [100, 100, 100, 100, 100],
        ...         ]
        ...     )

        Now let's create an ElevationRaster from this array with a resolution of 10.

        >>> r = sg.ElevationRaster.from_array(arr, crs=None, bounds=(0, 0, 50, 50))
        >>> r.res
        (10.0, 10.0)

        The gradient will be 1 (1 meter up for every meter forward).
        The calculation is by default done in place to save memory.

        >>> r.gradient()
        >>> r.array
        array([[0., 1., 1., 1., 0.],
            [1., 1., 1., 1., 1.],
            [1., 1., 0., 1., 1.],
            [1., 1., 1., 1., 1.],
            [0., 1., 1., 1., 0.]])
        """
        if len(self.array.shape) == 2:
            array = self._slope_2d(self.array, degrees=degrees)
        else:
            out_array = []
            for array in self.array:
                results = self._slope_2d(array, degrees=degrees)
                out_array.append(results)
            array = np.array(out_array)

        return self._return_self_or_copy(array, copy)

    def _slope_2d(self, array, degrees) -> np.ndarray:
        gradient_x, gradient_y = np.gradient(array, self.res[0], self.res[1])

        gradient = abs(gradient_x) + abs(gradient_y)

        if not degrees:
            return gradient

        radians = np.arctan(gradient)
        degrees = np.degrees(radians)

        assert np.max(degrees) <= 90

        return degrees



"""Interactive or static map of one or more GeoDataFrames.

This module holds the Map class, which is the basis for the Explore class.
"""
import warnings

import matplotlib
import matplotlib.colors as colors
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame, GeoSeries
from jenkspy import jenks_breaks
from mapclassify import classify
from shapely import Geometry

from ..geopandas_tools.conversion import to_gdf
from ..geopandas_tools.general import (
    clean_geoms,
    drop_inactive_geometry_columns,
    get_common_crs,
    rename_geometry_if,
)
from ..helpers import get_object_name


# the geopandas._explore raises a deprication warning. Ignoring for now.
warnings.filterwarnings(
    action="ignore", category=matplotlib.MatplotlibDeprecationWarning
)
pd.options.mode.chained_assignment = None


# custom default colors for non-numeric data, because the geopandas default has very
# similar colors. The palette is like the "Set2" cmap from matplotlib, but with more
# colors. If more than 14 categories, the geopandas default cmap is used.
_CATEGORICAL_CMAP = {
    0: "#4576ff",
    1: "#ff455e",
    2: "#ffa617",
    3: "#ff8cc9",
    4: "#804e00",
    5: "#99ff00",
    6: "#fff700",
    7: "#00ffee",
    8: "#36d19b",
    9: "#94006b",
    10: "#750000",
    11: "#1c6b00",
}

DEFAULT_SCHEME = "quantiles"


def proper_fillna(val, fill_val):
    """fillna doesn't always work. So doing it manually."""
    try:
        if "NAType" in val.__class__.__name__:
            return fill_val
    except Exception:
        if fill_val is None:
            return fill_val
        if fill_val != fill_val:
            return fill_val

    return val


class Map:
    """Base class that prepares one or more GeoDataFrames for mapping.

    The class has no public methods.
    """

    def __init__(
        self,
        *gdfs: GeoDataFrame,
        column: str | None = None,
        labels: tuple[str] | None = None,
        k: int = 5,
        bins: tuple[float] | None = None,
        nan_label: str = "Missing",
        nan_color="#c2c2c2",
        scheme: str = DEFAULT_SCHEME,
        **kwargs,
    ):
        gdfs, column, kwargs = self._separate_args(gdfs, column, kwargs)

        self._column = column
        self.bins = bins
        self._k = k
        self.nan_label = nan_label
        self.nan_color = nan_color
        self._cmap = kwargs.pop("cmap", None)
        self.scheme = scheme

        if not all(isinstance(gdf, GeoDataFrame) for gdf in gdfs):
            gdfs = [
                to_gdf(gdf) if not isinstance(gdf, GeoDataFrame) else gdf
                for gdf in gdfs
            ]
            if not all(isinstance(gdf, GeoDataFrame) for gdf in gdfs):
                raise ValueError("gdfs must be GeoDataFrames.")

        if "namedict" in kwargs:
            for i, gdf in enumerate(gdfs):
                gdf.name = kwargs["namedict"][i]
            kwargs.pop("namedict")

        # need to get the object names of the gdfs before copying. Only getting,
        # not setting, labels. So the original gdfs don't get the label column.
        self.labels = labels
        if not self.labels:
            self._get_labels(gdfs)

        show = kwargs.pop("show", True)
        if isinstance(show, (int, bool)):
            show_temp = [bool(show) for _ in range(len(gdfs))]
        elif not hasattr(show, "__iter__") or len(show) != len(gdfs):
            raise ValueError(
                "'show' must be boolean or an iterable of boleans same length as gdfs"
            )
        else:
            show_temp = show

        self._gdfs = []
        new_labels = []
        self.show = []
        for label, gdf, show in zip(self.labels, gdfs, show_temp, strict=True):
            if not len(gdf):
                continue

            gdf = clean_geoms(gdf).reset_index(drop=True)
            if not len(gdf):
                continue

            self._gdfs.append(gdf)
            new_labels.append(label)
            self.show.append(show)
        self.labels = new_labels

        if len(self._gdfs):
            last_show = self.show[-1]
        else:
            last_show = show

        # pop all geometry-like items from kwargs into self._gdfs
        self.kwargs = {}
        for key, value in kwargs.items():
            if isinstance(value, GeoDataFrame):
                self._gdfs.append(value)
                self.labels.append(key)
                self.show.append(last_show)
                continue
            try:
                self._gdfs.append(to_gdf(value))
                self.labels.append(key)
                self.show.append(last_show)
            except Exception:
                self.kwargs[key] = value

        if not any(len(gdf) for gdf in self._gdfs):
            warnings.warn("None of the GeoDataFrames have rows.")
            self._gdfs = None
            self._is_categorical = True
            self._unique_values = []
            return

        if not self.labels:
            self._set_labels()

        self._gdfs = self._to_common_crs_and_one_geom_col(self._gdfs)
        self._is_categorical = self._check_if_categorical()

        if self._column:
            self._fillna_if_col_is_missing()
        else:
            gdfs = []
            for gdf, label in zip(self._gdfs, self.labels, strict=True):
                gdf["label"] = label
                gdfs.append(gdf)
            self._column = "label"
            self._gdfs = gdfs

        try:
            self._gdf = pd.concat(self._gdfs, ignore_index=True)
        except ValueError:
            crs = get_common_crs(self._gdfs)
            for gdf in self._gdfs:
                gdf.crs = crs
            self._gdf = pd.concat(self._gdfs, ignore_index=True)

        self._nan_idx = self._gdf[self._column].isna()
        self._get_unique_values()

    def _get_unique_values(self):
        if not self._is_categorical:
            self._unique_values = self._get_unique_floats()
        else:
            unique = list(self._gdf[self._column].unique())
            try:
                self._unique_values = sorted(unique)
            except TypeError:
                self._unique_values = [
                    x for x in unique if proper_fillna(x, None) is not None
                ]

        self._k = min(self._k, len(self._unique_values))

    def _get_unique_floats(self) -> np.array:
        """Get unique floats by multiplying, then converting to integer.

        Find a multiplier that makes the max value greater than +- 1_000_000.
        Because floats don't always equal each other. This will make very
        similar values count as the same value in the color classification.
        """
        array = self._gdf.loc[~self._nan_idx, self._column]
        self._min = np.min(array)
        self._max = np.max(array)
        self._get_multiplier(array)

        unique = array.reset_index(drop=True).drop_duplicates()
        as_int = self._array_to_large_int(unique)
        no_duplicates = as_int.drop_duplicates()

        return np.sort(np.array(unique.loc[no_duplicates.index]))

    def _array_to_large_int(self, array):
        """Multiply values in float array, then convert to integer."""
        if not isinstance(array, pd.Series):
            array = pd.Series(array)

        notna = array[array.notna()]
        isna = array[array.isna()]

        unique_multiplied = (notna * self._multiplier).astype(np.int64)

        return pd.concat([unique_multiplied, isna]).sort_index()

    def _get_multiplier(self, array: np.ndarray):
        """Find the number of zeros needed to push the max value of the array above
        +-1_000_000.

        Adding this as an attribute to use later in _classify_from_bins.
        """
        if np.max(array) == 0:
            self._multiplier: int = 1
            return

        multiplier = 10
        max_ = np.max(array * multiplier)

        if self._max > 0:
            while max_ < 1_000_000:
                multiplier *= 10
                max_ = np.max(array * multiplier)
        else:
            while max_ > -1_000_000:
                multiplier *= 10
                max_ = np.max(array * multiplier)

        self._multiplier: int = multiplier

    def _add_minmax_to_bins(self, bins: list[float | int]) -> list[float | int]:
        """If values are outside the bin range, add max and/or min values of array."""
        # make sure they are lists
        bins = [bin for bin in bins]

        if min(bins) > 0 and min(self._gdf.loc[~self._nan_idx, self._column]) < min(
            bins
        ):
            bins = [min(self._gdf.loc[~self._nan_idx, self._column])] + bins

        if min(bins) < 0 and min(self._gdf.loc[~self._nan_idx, self._column]) < min(
            bins
        ):
            bins = [min(self._gdf.loc[~self._nan_idx, self._column])] + bins

        if max(bins) > 0 and max(
            self._gdf.loc[self._gdf[self._column].notna(), self._column]
        ) > max(bins):
            bins = bins + [
                max(self._gdf.loc[self._gdf[self._column].notna(), self._column])
            ]

        if max(bins) < 0 and max(
            self._gdf.loc[self._gdf[self._column].notna(), self._column]
        ) < max(bins):
            bins = bins + [
                max(self._gdf.loc[self._gdf[self._column].notna(), self._column])
            ]

        return bins

    @staticmethod
    def _separate_args(
        args: tuple,
        column: str | None,
        kwargs: dict,
    ) -> tuple[tuple[GeoDataFrame], str]:
        """Separate GeoDataFrames from string (column argument)."""

        def as_dict(obj):
            if hasattr(obj, "__dict__"):
                return obj.__dict__
            elif isinstance(obj, dict):
                return obj
            raise TypeError

        gdfs: tuple[GeoDataFrame] = ()
        for arg in args:
            if isinstance(arg, str):
                if column is None:
                    column = arg
                else:
                    raise ValueError(
                        "Can specify at most one string as a positional argument."
                    )
            elif isinstance(arg, (GeoDataFrame, GeoSeries, Geometry)):
                gdfs = gdfs + (arg,)
            elif isinstance(arg, dict) or hasattr(arg, "__dict__"):
                # add dicts or classes with GeoDataFrames to kwargs
                more_gdfs = {}
                for key, value in as_dict(arg).items():
                    if isinstance(value, (GeoDataFrame, GeoSeries, Geometry)):
                        more_gdfs[key] = value
                    elif isinstance(value, dict) or hasattr(value, "__dict__"):
                        try:
                            # same as above, one level down
                            more_gdfs |= {
                                k: v
                                for k, v in value.items()
                                if isinstance(v, (GeoDataFrame, GeoSeries, Geometry))
                            }
                        except Exception:
                            # no need to raise here
                            pass

                kwargs |= more_gdfs

        return gdfs, column, kwargs

    def _prepare_continous_map(self):
        """Create bins if not already done and adjust k if needed."""

        if self.scheme is None:
            return

        if not self.bins:
            self.bins = self._create_bins(self._gdf, self._column)
            if len(self.bins) <= self._k and len(self.bins) != len(self._unique_values):
                self._k = len(self.bins)
        elif not all(self._gdf[self._column].isna()):
            self.bins = self._add_minmax_to_bins(self.bins)
            if len(self._unique_values) <= len(self.bins):
                self._k = len(self.bins)  # - 1
        else:
            self._unique_values = self.nan_label
            self._k = 1

    def _get_labels(self, gdfs: tuple[GeoDataFrame]) -> None:
        """Putting the labels/names in a list before copying the gdfs."""
        self.labels: list[str] = []
        for i, gdf in enumerate(gdfs):
            if hasattr(gdf, "name") and isinstance(gdf.name, str):
                name = gdf.name
            else:
                name = get_object_name(gdf)
                name = name or str(i)
            self.labels.append(name)

    def _set_labels(self) -> None:
        """Setting the labels after copying the gdfs."""
        gdfs = []
        for i, gdf in enumerate(self._gdfs):
            gdf["label"] = self.labels[i]
            gdfs.append(gdf)
        self._gdfs = gdfs

    def _to_common_crs_and_one_geom_col(self, gdfs: list[GeoDataFrame]):
        """Need common crs and max one geometry column."""
        crs_list = list({gdf.crs for gdf in gdfs if gdf.crs is not None})
        if crs_list:
            self.crs = crs_list[0]
        new_gdfs = []
        for gdf in gdfs:
            gdf = gdf.reset_index(drop=True)
            gdf = drop_inactive_geometry_columns(gdf).pipe(rename_geometry_if)
            if crs_list:
                try:
                    gdf = gdf.to_crs(self.crs)
                except ValueError:
                    gdf = gdf.set_crs(self.crs)
            new_gdfs.append(gdf)
        return new_gdfs

    def _fillna_if_col_is_missing(self) -> None:
        n = 0
        for gdf in self._gdfs:
            if self._column in gdf.columns:
                gdf[self._column] = gdf[self._column].fillna(pd.NA)
                n += 1
            else:
                gdf[self._column] = pd.NA

        maybe_area = 1 if "area" in self._column else 0
        maybe_length = (
            1 if any(x in self._column for x in ["meter", "metre", "leng"]) else 0
        )
        n = n + maybe_area + maybe_length

        if n == 0:
            raise ValueError(
                f"The column {self._column!r} is not present in any "
                "of the passed GeoDataFrames."
            )

    def _check_if_categorical(self) -> bool:
        """Quite messy this..."""
        if not self._column:
            return True

        maybe_area = 1 if "area" in self._column else 0
        maybe_length = (
            1 if any(x in self._column for x in ["meter", "metre", "leng"]) else 0
        )

        all_nan = 0
        col_not_present = 0
        for gdf in self._gdfs:
            if self._column not in gdf:
                if maybe_area:
                    gdf["area"] = gdf.area
                    maybe_area += 1
                elif maybe_length:
                    gdf["length"] = gdf.length
                    maybe_length += 1
                else:
                    col_not_present += 1
            elif not pd.api.types.is_numeric_dtype(gdf[self._column]):
                if all(gdf[self._column].isna()):
                    all_nan += 1
                return True

        if maybe_area > 1:
            self._column = "area"
            return False
        if maybe_length > 1:
            self._column = "length"
            return False

        if all_nan == len(self._gdfs):
            raise ValueError(f"All values are NaN in column {self.column!r}.")

        if col_not_present == len(self._gdfs):
            raise ValueError(f"{self.column} not found.")

        return False

    def _get_categorical_colors(self) -> None:
        # custom categorical cmap
        if not self._cmap and len(self._unique_values) <= len(_CATEGORICAL_CMAP):
            self._categories_colors_dict = {
                category: _CATEGORICAL_CMAP[i]
                for i, category in enumerate(self._unique_values)
            }
        elif self._cmap:
            cmap = matplotlib.colormaps.get_cmap(self._cmap)

            self._categories_colors_dict = {
                category: colors.to_hex(cmap(int(i)))
                for i, category in enumerate(self._unique_values)
            }
        else:
            cmap = matplotlib.colormaps.get_cmap("tab20")

            self._categories_colors_dict = {
                category: colors.to_hex(cmap(int(i)))
                for i, category in enumerate(self._unique_values)
            }

        if any(self._nan_idx):
            self._gdf[self._column] = self._gdf[self._column].fillna(self.nan_label)
            self._categories_colors_dict[self.nan_label] = self.nan_color

        new_gdfs = []
        for gdf in self._gdfs:
            gdf["color"] = gdf[self._column].map(self._categories_colors_dict)
            new_gdfs.append(gdf)
        self._gdfs = new_gdfs

        self._gdf["color"] = self._gdf[self._column].map(self._categories_colors_dict)

    def _create_bins(self, gdf: GeoDataFrame, column: str) -> np.ndarray:
        """Make bin list of length k + 1, or length of unique values.

        The returned bins sometimes have two almost identical

        If 'scheme' is not specified, the jenks_breaks function is used, which is
        much faster than the one from Mapclassifier.
        """

        if not len(gdf.loc[~self._nan_idx, column]):
            return np.array([0])

        n_classes = (
            self._k if len(self._unique_values) > self._k else len(self._unique_values)
        )

        if self._k == len(self._unique_values) - 1:
            n_classes = self._k - 1
            self._k = self._k - 1

        if self._k > len(self._unique_values):
            self._k = len(self._unique_values)
            n_classes = len(self._unique_values)

        if self.scheme == "jenks":
            try:
                bins = jenks_breaks(
                    gdf.loc[~self._nan_idx, column], n_classes=n_classes
                )
                bins = self._add_minmax_to_bins(bins)
            except Exception:
                pass
        else:
            binning = classify(
                np.asarray(gdf.loc[~self._nan_idx, column]),
                scheme=self.scheme,
                k=self._k,
            )
            bins = binning.bins
            bins = self._add_minmax_to_bins(bins)

        unique_bins = list({round(bin, 5) for bin in bins})
        unique_bins.sort()

        if self._k == len(self._unique_values) - 1:
            return np.array(unique_bins)

        if len(unique_bins) == len(self._unique_values):
            return np.array(unique_bins)

        if len(unique_bins) == len(bins) - 1:
            self._k -= 1

        return np.array(bins)

    def change_cmap(self, cmap: str, start: int = 0, stop: int = 256):
        """Change the color palette of the plot.

        Args:
            cmap: The colormap.
                https://matplotlib.org/stable/tutorials/colors/colormaps.html
            start: Start position for the color palette. Defaults to 0.
            stop: End position for the color palette. Defaults to 256, which
                is the end of the color range.
        """
        self.cmap_start = start
        self.cmap_stop = stop
        self._cmap = cmap
        self._cmap_has_been_set = True
        return self

    def _get_continous_colors(self, n: int) -> np.ndarray:
        cmap = matplotlib.colormaps.get_cmap(self._cmap)
        colors_ = [
            colors.to_hex(cmap(int(i)))
            #            for i in np.linspace(self.cmap_start, self.cmap_stop, num=self._k)
            for i in np.linspace(self.cmap_start, self.cmap_stop, num=n)
        ]
        if any(self._nan_idx):
            colors_ = colors_ + [self.nan_color]
        return np.array(colors_)

    def _classify_from_bins(self, gdf: GeoDataFrame, bins: np.ndarray) -> np.ndarray:
        """Place the column values into groups."""

        # if equal lenght, convert to integer and check for equality
        if len(bins) == len(self._unique_values):
            if gdf[self._column].isna().all():
                return np.repeat(len(bins), len(gdf))

            gdf["col_as_int"] = self._array_to_large_int(gdf[self._column])
            bins = self._array_to_large_int(self._unique_values)
            gdf["col_as_int"] = gdf["col_as_int"].fillna(np.nan)
            classified = np.searchsorted(bins, gdf["col_as_int"])

        else:
            if len(bins) == self._k + 1:
                bins = bins[1:]

            if gdf[self._column].isna().all():
                return np.repeat(len(bins), len(gdf))

            # need numpy.nan instead of pd.NA
            gdf[self._column] = gdf[self._column].fillna(np.nan)

            gdf[self._column] = gdf[self._column].apply(
                lambda x: proper_fillna(x, np.nan)
            )

            classified = np.searchsorted(bins, gdf[self._column])

        return classified

    def _push_classification(self, classified: np.ndarray) -> np.ndarray:
        """Push classes downwards if gaps in classification sequence.

        So from e.g. [0,2,4] to [0,1,2].

        Otherwise, will get index error when classifying colors.
        """
        rank_dict = {val: rank for rank, val in enumerate(np.unique(classified))}

        return np.array([rank_dict[val] for val in classified])

    @property
    def k(self):
        return self._k

    @k.setter
    def k(self, new_value: bool):
        if not self._is_categorical and new_value > len(self._unique_values):
            raise ValueError(
                "'k' cannot be greater than the number of unique values in the column.'"
                f"Got new k={new_value} and previous k={len(self._unique_values)}.'"
                #  f"{''.join(self._unique_values)}"
            )
        self._k = int(new_value)

    @property
    def cmap(self):
        return self._cmap

    @cmap.setter
    def cmap(self, new_value: bool):
        self._cmap = new_value
        self.change_cmap(cmap=new_value, start=self.cmap_start, stop=self.cmap_stop)

    @property
    def gdf(self):
        return self._gdf

    @gdf.setter
    def gdf(self, _):
        raise ValueError(
            "Cannot change 'gdf' after init. Put the GeoDataFrames into "
            "the class initialiser."
        )

    @property
    def gdfs(self):
        return self._gdfs

    @gdfs.setter
    def gdfs(self, _):
        raise ValueError(
            "Cannot change 'gdfs' after init. Put the GeoDataFrames into "
            "the class initialiser."
        )

    @property
    def column(self):
        return self._column

    @column.setter
    def column(self, _):
        raise ValueError(
            "Cannot change 'column' after init. Specify 'column' in the "
            "class initialiser."
        )

    def __setitem__(self, item, new_item):
        return setattr(self, item, new_item)

    def __getitem__(self, item):
        return getattr(self, item)

    def get(self, key, default=None):
        try:
            return self[key]
        except (KeyError, ValueError, IndexError, AttributeError):
            return default



import numpy as np

from .raster import Raster


class Sentinel2(Raster):
    band_colors = {
        "B1": "coastal aerosol",
        "B2": "blue",
        "B3": "green",
        "B4": "red",
        "B5": "vegetation red edge",
        "B6": "vegetation red edge",
        "B7": "vegetation red edge",
        "B8": "nir",
        "B8A": "narrow nir",
        "B9": "water vapour",
        "B10": "swir - cirrus",
        "B11": "swir",
        "B12": "swir",
    }

    def __init__(self, raster=None, **kwargs):
        kwargs = {
            "nodata": 0,
            "dtype": np.uint16,
            "band_index": 1,
            "name_regex": r"B\d{1,2}A|B\d{1,2}",
            "date_regex": r"20\d{6}",
            "shortname": "sentinel2",
        } | kwargs

        super().__init__(raster, **kwargs)

    @property
    def band_color(self):
        if not self.name:
            return None
        return self.colors[self.name]

    @property
    def is_mask(self):
        return "masks" in str(self.path).lower()



"""Cutting and splitting line geometries."""
import warnings

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame, GeoSeries
from pandas import DataFrame, Series
from shapely import (
    buffer,
    extract_unique_points,
    force_2d,
    get_coordinates,
    get_parts,
    linestrings,
    touches,
    unary_union,
)
from shapely.geometry import LineString, Point

from ..geopandas_tools.buffer_dissolve_explode import buff
from ..geopandas_tools.conversion import to_gdf
from ..geopandas_tools.geometry_types import get_geom_type
from ..geopandas_tools.neighbors import get_k_nearest_neighbors
from ..geopandas_tools.point_operations import (
    _shapely_snap,
    snap_all,
    snap_within_distance,
)
from ..geopandas_tools.sfilter import sfilter_split
from .nodes import make_edge_coords_cols


def split_lines_by_nearest_point(
    gdf: GeoDataFrame,
    points: GeoDataFrame,
    max_distance: int | float | None = None,
    splitted_col: str | None = None,
) -> DataFrame:
    """Split lines that are closest to s point.

    Snaps points to nearest lines and splits the lines in two at the snap point.
    The splitting is done pointwise, meaning each point splits one line in two.
    The line will not be split if the point is closest to the endpoint of the line.

    This function is used in NetworkAnalysis if split_lines is set to True.

    Args:
        gdf: GeoDataFrame of lines that will be split.
        points: GeoDataFrame of points to split the lines with.
        max_distance: the maximum distance between the point and the line. Points
            further away than max_distance will not split any lines. Defaults to None.
        splitted_col: Optionally add a column

    Returns:
        A GeoDataFrame with the same columns as the input lines, but with the lines
        split at the closest point to the points. If no lines are within 'max_distance'
        from any points, 'gdf' is returned as it came.

    Raises:
        ValueError: If the crs of the input data differs.

    Examples
    --------
    >>> from sgis import read_parquet_url, split_lines_by_nearest_point
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")
    >>> points = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_oslo.parquet")
    >>> rows = len(roads)
    >>> rows
    93395

    Splitting lines for points closer than 10 meters from the lines.

    >>> roads = split_lines_by_nearest_point(roads, points, max_distance=10)
    >>> print("number of lines that were split:", len(roads) - rows)
    number of lines that were split: 380

    Splitting lines by all points.

    >>> roads = split_lines_by_nearest_point(roads, points)
    >>> print("number of lines that were split:", len(roads) - rows)
    number of lines that were split: 848

    Not all lines were split. That is because some points were closest to an endpoint
    of a line.
    """
    PRECISION = 1e-6

    if not len(gdf):
        return gdf

    if (points.crs is not None and gdf.crs is not None) and not points.crs.equals(
        gdf.crs
    ):
        raise ValueError("crs mismatch:", points.crs, "and", gdf.crs)

    if get_geom_type(gdf) != "line":
        raise ValueError("'gdf' should only have line geometries.", gdf.geom_type)

    if get_geom_type(points) != "point":
        raise ValueError("'points' should only have point geometries.")

    gdf = gdf.copy()

    # move the points to the nearest exact location on the line
    if max_distance:
        snapped = snap_within_distance(points, gdf, max_distance=max_distance)
    else:
        snapped = snap_all(points, gdf)

    # find the lines that were snapped to (or are very close because of float rounding)
    snapped_buff = buff(snapped, PRECISION, resolution=16)
    relevant_lines, the_other_lines = sfilter_split(gdf, snapped_buff)

    if max_distance and not len(relevant_lines):
        if splitted_col:
            return gdf.assign(**{splitted_col: 1})
        return gdf

    # need consistent coordinate dimensions later
    # (doing it down here to not overwrite the original data)
    relevant_lines.geometry = force_2d(relevant_lines.geometry)
    snapped.geometry = force_2d(snapped.geometry)

    # split the lines with buffer + difference, since shaply.split usually doesn't work
    # relevant_lines["_idx"] = range(len(relevant_lines))
    splitted = relevant_lines.overlay(snapped_buff, how="difference").explode(
        ignore_index=True
    )

    # linearrings (maybe coded as linestrings) that were not split,
    # do not have edges and must be added in the end
    boundaries = splitted.geometry.boundary
    circles = splitted[boundaries.is_empty]
    splitted = splitted[~boundaries.is_empty]

    if not len(splitted):
        return pd.concat([the_other_lines, circles], ignore_index=True)

    # the endpoints of the new lines are now sligtly off. Using get_k_nearest_neighbors
    # to get the exact snapped point coordinates, . This will map the sligtly
    # wrong line endpoints with the point the line was split by.

    snapped["point_coords"] = [(geom.x, geom.y) for geom in snapped.geometry]

    # get line endpoints as columns (source_coords and target_coords)
    splitted = make_edge_coords_cols(splitted)

    splitted_source = to_gdf(splitted["source_coords"], crs=gdf.crs)
    splitted_target = to_gdf(splitted["target_coords"], crs=gdf.crs)

    def get_nearest(splitted, snapped) -> pd.DataFrame:
        """find the nearest snapped point for each source and target of the lines"""
        return get_k_nearest_neighbors(splitted, snapped, k=1).loc[
            lambda x: x["distance"] <= PRECISION * 2
        ]

    # snapped = snapped.set_index("point_coords")
    snapped.index = snapped.geometry
    dists_source = get_nearest(splitted_source, snapped)
    dists_target = get_nearest(splitted_target, snapped)

    # neighbor_index: point coordinates as tuple
    pointmapper_source: pd.Series = dists_source["neighbor_index"]
    pointmapper_target: pd.Series = dists_target["neighbor_index"]

    # now, we can replace the source/target coordinate with the coordinates of
    # the snapped points.

    splitted = change_line_endpoint(
        splitted,
        indices=dists_source.index,
        pointmapper=pointmapper_source,
        change_what="first",
    )  # i=0)

    # same for the lines where the target was split, but change the last coordinate
    splitted = change_line_endpoint(
        splitted,
        indices=dists_target.index,
        pointmapper=pointmapper_target,
        change_what="last",
    )  # , i=-1)

    if splitted_col:
        splitted[splitted_col] = 1

    return pd.concat([the_other_lines, splitted, circles], ignore_index=True).drop(
        ["source_coords", "target_coords"], axis=1
    )


def change_line_endpoint(
    gdf: GeoDataFrame,
    indices: pd.Index,
    pointmapper: pd.Series,
    change_what: str | int,
) -> GeoDataFrame:
    """
    Loop for each line where the source is the endpoint that was split
    change the first point of the line to the point it was split by
    """
    assert gdf.index.is_unique

    if change_what == "first" or change_what == 0:
        to_be_changed = lambda x: ~x.index.duplicated(keep="first")
    elif change_what == "last" or change_what == -1:
        to_be_changed = lambda x: ~x.index.duplicated(keep="last")
    else:
        raise ValueError(
            f"change_what should be 'first' or 'last' or 0 or -1. Got {change_what}"
        )

    is_relevant = gdf.index.isin(indices)
    relevant_lines = gdf.loc[is_relevant]

    relevant_lines.geometry = extract_unique_points(relevant_lines.geometry)
    relevant_lines = relevant_lines.explode(index_parts=False)

    relevant_lines.loc[to_be_changed, "geometry"] = (
        relevant_lines.loc[to_be_changed].index.map(pointmapper).values
    )

    relevant_lines_mapped = relevant_lines.groupby(level=0)["geometry"].agg(LineString)

    gdf.loc[is_relevant, "geometry"] = relevant_lines_mapped

    return gdf


def cut_lines(gdf: GeoDataFrame, max_length: int, ignore_index=False) -> GeoDataFrame:
    """Cuts lines of a GeoDataFrame into pieces of a given length.

    Args:
        gdf: GeoDataFrame.
        max_length: The maximum length of the lines in the output GeoDataFrame.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False.

    Returns:
        A GeoDataFrame with lines cut to the maximum distance.

    Note:
        This method is time consuming for large networks and low 'max_length'.

    Examples
    --------
    >>> from sgis import read_parquet_url, cut_lines
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")
    >>> roads.length.describe().round(1)
    count    93395.0
    mean        41.2
    std         78.5
    min          0.2
    25%         14.0
    50%         27.7
    75%         47.5
    max       5213.7
    dtype: float64
    >>> roads = cut_lines(roads, max_length=100)
    >>> roads.length.describe().round(1)
    count    126304.0
    mean         30.5
    std          30.1
    min           0.0
    25%           5.7
    50%          22.5
    75%          44.7
    max         100.0
    dtype: float64
    """
    if get_geom_type(gdf) != "line":
        raise ValueError("'gdf' should only have line geometries.")

    gdf["geometry"] = force_2d(gdf.geometry)

    gdf = gdf.explode(ignore_index=ignore_index, index_parts=False)

    long_lines = gdf.loc[gdf.length > max_length]

    if not len(long_lines):
        return gdf

    for x in [10, 5, 1]:
        max_ = max(long_lines.length)
        while max_ > max_length * x + 1:
            max_ = max(long_lines.length)

            long_lines = cut_lines_once(long_lines, max_length)

            if max_ == max(long_lines.length):
                break

    long_lines = long_lines.explode(ignore_index=ignore_index, index_parts=False)

    short_lines = gdf.loc[gdf.length <= max_length]

    return pd.concat([short_lines, long_lines], ignore_index=ignore_index)


def cut_lines_once(
    gdf: GeoDataFrame,
    distances: int | float | str | Series,
    ignore_index: bool = False,
) -> GeoDataFrame:
    """Cuts lines of a GeoDataFrame in two at the given distance or distances.

    Takes a GeoDataFrame of lines and cuts each line in two. If distances is a number,
    all lines will be cut at the same length.

    Args:
        gdf: GeoDataFrame.
        distances: The distance from the start of the lines to cut at. Either a number,
            the name of a column or array-like of same length as the line GeoDataFrame.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False.

    Examples
    --------
    >>> from sgis import cut_lines_once, to_gdf
    >>> import pandas as pd
    >>> from shapely.geometry import LineString
    >>> gdf = to_gdf(LineString([(0, 0), (1, 1), (2, 2)]))
    >>> gdf = pd.concat([gdf, gdf, gdf])
    >>> gdf
                                                geometry
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...

    Cut all lines at 1 unit from the start of the lines.

    >>> cut_lines_once(gdf, 1)
                                                geometry
    0      LINESTRING (0.00000 0.00000, 0.70711 0.70711)
    1  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...
    2      LINESTRING (0.00000 0.00000, 0.70711 0.70711)
    3  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...
    4      LINESTRING (0.00000 0.00000, 0.70711 0.70711)
    5  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...

    Cut distance as column.

    >>> gdf["dist"] = [1, 2, 3]
    >>> cut_lines_once(gdf, "dist")
                                                geometry  dist
    0      LINESTRING (0.00000 0.00000, 0.70711 0.70711)     1
    1  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...     1
    2  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     2
    3      LINESTRING (1.41421 1.41421, 2.00000 2.00000)     2
    4  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     3

    Cut distance as list (same result as above).

    >>> cut_lines_once(gdf, [1, 2, 3])
                                                geometry  dist
    0      LINESTRING (0.00000 0.00000, 0.70711 0.70711)     1
    1  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...     1
    2  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     2
    3      LINESTRING (1.41421 1.41421, 2.00000 2.00000)     2
    4  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     3
    """

    def _cut(line: LineString, distance: int | float) -> list[LineString]:
        """From the shapely docs"""
        if distance <= 0.0 or distance >= line.length:
            return line
        coords = list(line.coords)
        for i, p in enumerate(coords):
            prd = line.project(Point(p))
            if prd == distance:
                return [LineString(coords[: i + 1]), LineString(coords[i:])]
            if prd > distance:
                cp = line.interpolate(distance)
                return [
                    LineString(coords[:i] + [(cp.x, cp.y)]),
                    LineString([(cp.x, cp.y)] + coords[i:]),
                ]

    crs = gdf.crs
    geom_col = gdf._geometry_column_name

    gdf = gdf.copy()

    # cutting lines will give lists of linestrings in the geometry column. Ignoring
    # the warning it triggers
    warnings.filterwarnings(
        "ignore", message="Geometry column does not contain geometry."
    )

    if isinstance(distances, str):
        gdf[geom_col] = np.vectorize(_cut)(gdf[geom_col], gdf[distances])
    else:
        gdf[geom_col] = np.vectorize(_cut)(gdf[geom_col], distances)

    # explode will give pandas df if not gdf is constructed
    gdf = GeoDataFrame(
        gdf.explode(ignore_index=ignore_index, index_parts=False),
        geometry=geom_col,
        crs=crs,
    )

    return gdf



from .geopandas_tools.bounds import (
    Gridlooper,
    bounds_to_points,
    bounds_to_polygon,
    get_total_bounds,
    gridloop,
    make_grid,
    make_grid_from_bbox,
    make_ssb_grid,
    points_in_bounds,
    to_bbox,
)
from .geopandas_tools.buffer_dissolve_explode import (
    buff,
    buffdiss,
    buffdissexp,
    buffdissexp_by_cluster,
    dissexp,
    dissexp_by_cluster,
)
from .geopandas_tools.centerlines import get_rough_centerlines
from .geopandas_tools.cleaning import (
    coverage_clean,
    remove_spikes,
    split_spiky_polygons,
)
from .geopandas_tools.conversion import (
    coordinate_array,
    from_4326,
    to_4326,
    to_gdf,
    to_geoseries,
    to_shapely,
)
from .geopandas_tools.duplicates import (  # drop_duplicate_geometries,
    get_intersections,
    update_geometries,
)
from .geopandas_tools.general import (
    clean_clip,
    clean_geoms,
    drop_inactive_geometry_columns,
    get_common_crs,
    get_grouped_centroids,
    random_points,
    random_points_in_polygons,
    rename_geometry_if,
    sort_large_first,
    sort_long_first,
    sort_short_first,
    sort_small_first,
    to_lines,
)
from .geopandas_tools.geocoding import address_to_coords, address_to_gdf
from .geopandas_tools.geometry_types import (
    get_geom_type,
    is_single_geom_type,
    make_all_singlepart,
    to_single_geom_type,
)
from .geopandas_tools.neighbors import (
    get_all_distances,
    get_k_nearest_neighbors,
    get_neighbor_dfs,
    get_neighbor_indices,
    k_nearest_neighbors,
)
from .geopandas_tools.overlay import clean_overlay
from .geopandas_tools.point_operations import snap_all, snap_within_distance
from .geopandas_tools.polygon_operations import (
    PolygonsAsRings,
    close_all_holes,
    close_small_holes,
    close_thin_holes,
    eliminate_by_largest,
    eliminate_by_longest,
    eliminate_by_smallest,
    get_gaps,
    get_holes,
    get_polygon_clusters,
)
from .geopandas_tools.polygons_as_rings import PolygonsAsRings
from .geopandas_tools.sfilter import sfilter, sfilter_inverse, sfilter_split
from .helpers import get_object_name, sort_nans_last
from .io.opener import opener
from .io.read_parquet import read_parquet_url
from .maps.examine import Examine
from .maps.explore import Explore
from .maps.httpserver import run_html_server
from .maps.legend import Legend
from .maps.maps import clipmap, explore, explore_locals, qtm, samplemap
from .maps.thematicmap import ThematicMap
from .maps.tilesources import kartverket as kartverket_tiles
from .maps.tilesources import xyz as xyztiles
from .networkanalysis.closing_network_holes import (
    close_network_holes,
    close_network_holes_to_deadends,
)
from .networkanalysis.cutting_lines import (
    cut_lines,
    cut_lines_once,
    split_lines_by_nearest_point,
)
from .networkanalysis.directednetwork import (
    make_directed_network,
    make_directed_network_norway,
)
from .networkanalysis.finding_isolated_networks import (
    get_component_size,
    get_connected_components,
)
from .networkanalysis.network import Network
from .networkanalysis.networkanalysis import NetworkAnalysis
from .networkanalysis.networkanalysisrules import NetworkAnalysisRules
from .networkanalysis.nodes import (
    make_edge_coords_cols,
    make_edge_wkt_cols,
    make_node_ids,
)
from .networkanalysis.traveling_salesman import traveling_salesman_problem
from .parallel.parallel import Parallel
from .raster.elevationraster import ElevationRaster
from .raster.raster import Raster
from .raster.sentinel import Sentinel2


try:
    from .io.dapla_functions import check_files, read_geopandas, write_geopandas
    from .io.write_municipality_data import write_municipality_data
except ImportError:
    pass



"""Check and set geometry type."""
import numpy as np
import pandas as pd
import shapely
from geopandas import GeoDataFrame, GeoSeries
from geopandas.array import GeometryArray
from shapely import Geometry


def make_all_singlepart(
    gdf: GeoDataFrame | GeoSeries, index_parts: bool = False, ignore_index: bool = False
) -> GeoDataFrame | GeoSeries:
    # only explode if nessecary
    if index_parts or ignore_index and not gdf.index.equals(pd.Index(range(len(gdf)))):
        gdf = gdf.explode(index_parts=index_parts, ignore_index=ignore_index)

    while not gdf.geom_type.isin(
        ["Polygon", "Point", "LineString", "LinearRing"]
    ).all():
        gdf = gdf.explode(index_parts=index_parts, ignore_index=ignore_index)

    return gdf


def to_single_geom_type(
    gdf: GeoDataFrame | GeoSeries,
    geom_type: str,
    ignore_index: bool = False,
) -> GeoDataFrame | GeoSeries:
    """Returns only the specified geometry type in a GeoDataFrame or GeoSeries.

    GeometryCollections are first exploded, then only the rows with the given
    geometry_type is kept. Both multipart and singlepart geometries are kept.
    LinearRings are considered lines.

    Args:
        gdf: GeoDataFrame or GeoSeries
        geom_type: the geometry type to keep, either 'point', 'line' or 'polygon'. Both
            multi- and singlepart geometries are included.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False.

    Returns:
        A GeoDataFrame with a single geometry type.

    Raises:
        TypeError: If incorrect gdf type.
        ValueError: If 'geom_type' is neither 'polygon', 'line' or 'point'.

    Examples
    --------
    First create a GeoDataFrame of mixed geometries.

    >>> from sgis import to_gdf, to_single_geom_type
    >>> from shapely.geometry import LineString, Polygon
    >>> gdf = to_gdf([
    ...     (0, 0),
    ...     LineString([(1, 1), (2, 2)]),
    ...     Polygon([(3, 3), (4, 4), (3, 4), (3, 3)])
    ...     ])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....

    >>> to_single_geom_type(gdf, "line")
                                            geometry
    1  LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    >>> to_single_geom_type(gdf, "polygon")
                                                geometry
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....

    Also keeps multigeometries and geometries within GeometryCollections.

    >>> gdf = gdf.dissolve()
    >>> gdf
                                                geometry
    0  GEOMETRYCOLLECTION (POINT (0.00000 0.00000), L...

    >>> to_single_geom_type(gdf, "line")
                                            geometry
    2  LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    """
    if all(g not in geom_type for g in ["polygon", "line", "point"]):
        raise ValueError(
            f"Invalid geom_type {geom_type!r}. Should be 'polygon', 'line' or 'point'"
        )

    if isinstance(gdf, Geometry):
        return _shapely_to_single_geom_type(gdf, geom_type)

    if isinstance(gdf, (np.ndarray, GeometryArray)):
        arr = np.vectorize(_shapely_to_single_geom_type)(gdf, geom_type)
        return arr[~shapely.is_empty(arr)]

    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):
        raise TypeError(f"'gdf' should be GeoDataFrame or GeoSeries, got {type(gdf)}")

    # explode collections to single-typed geometries
    collections = gdf.geom_type == "GeometryCollection"
    if collections.any():
        collections = make_all_singlepart(gdf[collections], ignore_index=ignore_index)

        gdf = pd.concat([gdf, collections], ignore_index=ignore_index)

    if "poly" in geom_type:
        is_polygon = gdf.geom_type.isin(["Polygon", "MultiPolygon"])
        if not is_polygon.all():
            gdf = gdf.loc[is_polygon]
    elif "line" in geom_type:
        is_line = gdf.geom_type.isin(["LineString", "MultiLineString", "LinearRing"])
        if not is_line.all():
            gdf = gdf.loc[is_line]
    else:
        is_point = gdf.geom_type.isin(["Point", "MultiPoint"])
        if not is_point.all():
            gdf = gdf.loc[is_point]

    return gdf.reset_index(drop=True) if ignore_index else gdf


def _shapely_to_single_geom_type(geom, geom_type):
    parts = shapely.get_parts(geom)
    return shapely.unary_union(
        [part for part in parts if geom_type.lower() in part.geom_type.lower()]
    )


def get_geom_type(gdf: GeoDataFrame | GeoSeries) -> str:
    """Returns a string of the geometry type in a GeoDataFrame or GeoSeries.

    Args:
        gdf: GeoDataFrame or GeoSeries

    Returns:
        A string that is either "polygon", "line", "point", or "mixed".

    Raises:
        TypeError: If 'gdf' is not of type GeoDataFrame or GeoSeries.

    Examples
    --------
    >>> from sgis import to_gdf, get_geom_type
    >>> gdf = to_gdf([0, 0])
    >>> gdf
                      geometry
    0  POINT (0.00000 0.00000)
    >>> get_geom_type(gdf)
    'point'
    """
    polys = ["Polygon", "MultiPolygon", None]
    lines = ["LineString", "MultiLineString", "LinearRing", None]
    points = ["Point", "MultiPoint", None]
    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):
        if isinstance(gdf, Geometry):
            if gdf.geom_type in polys:
                return "polygon"
            if gdf.geom_type in lines:
                return "line"
            if gdf.geom_type in points:
                return "point"
            return "mixed"
        raise TypeError(f"'gdf' should be GeoDataFrame or GeoSeries, got {type(gdf)}")

    if (gdf.geom_type.isin(polys)).all():
        return "polygon"
    if (gdf.geom_type.isin(lines)).all():
        return "line"
    if (gdf.geom_type.isin(points)).all():
        return "point"
    return "mixed"


def is_single_geom_type(gdf: GeoDataFrame | GeoSeries) -> bool:
    """Returns True if all geometries in a GeoDataFrame are of the same type.

    The types are either polygon, line or point. Multipart and singlepart are
    considered the same type.

    Args:
        gdf: GeoDataFrame or GeoSeries

    Returns:
        True if all geometries are the same type, False if not.

    Raises:
        TypeError: If 'gdf' is not of type GeoDataFrame or GeoSeries.

    Examples
    --------
    >>> from sgis import to_gdf, get_geom_type
    >>> gdf = to_gdf([0, 0])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    >>> is_single_geom_type(gdf)
    True
    """
    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):
        raise TypeError(f"'gdf' should be GeoDataFrame or GeoSeries, got {type(gdf)}")

    if all(gdf.geom_type.isin(["Polygon", "MultiPolygon"])):
        return True
    if all(gdf.geom_type.isin(["LineString", "MultiLineString", "LinearRing"])):
        return True
    if all(gdf.geom_type.isin(["Point", "MultiPoint"])):
        return True

    return False



"""The NetworkAnalysisRules class sets the rules for the network analysis.

The class is to be used as the 'rules' parameter in the NetworkAnalysis
class.
"""
from copy import copy, deepcopy
from dataclasses import dataclass

from geopandas import GeoDataFrame

from ..helpers import unit_is_meters


@dataclass
class NetworkAnalysisRules:
    """Sets the rules for the network analysis.

    To be used as the 'rules' parameter in the NetworkAnalysis class.

    Args:
        weight: Either a column in the GeoDataFrame of the Network or
            'meters'/'metres'. A 'minutes' column can be created with the
            'make_directed_network' method of the DirectedNetwork class.
        directed: Whether the lines will be considered traversable in both directions.
        search_tolerance: distance to search for nodes in the network. Origins and
            destinations further away from the network than the search_tolerance will
            not find any paths. Defaults to 250.
        search_factor: number of meters and percent to add to the closest distance to a
            node when connecting origins and destinations to the network. Defaults to
            0, meaning only the closest node is used. If search_factor is 10 and the
            closest node is 1 meter away, paths will be created from the point and all
            nodes within 11.1 meters. If the closest node is 100 meters away, paths
            will be created with all nodes within 120 meters.

            It can be wise to set a higher search_factor only for the origins and
            destinations that are causing problems in a separate analysis run.
        split_lines: If True (default), the closest line  to each point will be
            split in two at the nearest excact point. The weight of the split lines
            are then adjusted to the new length. If False, the points will be
            connected to the endpoints of the network lines.
        nodedist_kmh: When using "minutes" as weight, this sets the speed in kilometers
            per hour for the edges between origins/destinations and the network nodes
            that connect them. Defaults to None, meaning 0 weight is added for the
            edges.
        nodedist_multiplier: When using "meters" as weight, this sets the weight for
            the edges between origins/destinations and the network nodes that connect
            them. Defaults to None, meaning 0 weight is added for these edges. If set
            to 1, the weight will be equal to the straigt line distance.

    Examples
    --------
    Read testdata.

    >>> import sgis as sg
    >>> roads = sg.read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")
    >>> points = sg.read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_oslo.parquet")

    Let's start by setting the default rules. 'weight' and 'directed' have no default
    values.

    >>> rules = sg.NetworkAnalysisRules(weight="minutes", directed=True)
    >>> directed_roads = sg.remove_isolated(roads).pipe(sg.make_directed_network_norway)
    >>> nwa = sg.NetworkAnalysis(network=directed_roads, rules=rules, detailed_log=False)
    >>> nwa
    NetworkAnalysis(
        network=DirectedNetwork(6364 km, percent_bidirectional=87),
        rules=NetworkAnalysisRules(weight=minutes, search_tolerance=250, search_factor=0, split_lines=False, ...),
        log=True, detailed_log=True,
    )

    Setting 'split_lines' to True, means the points will be connected to the closest
    part of the closest network line. If False, the lines are connected to the closest
    endpoint of the lines. split_lines defaults to False, since splitting lines takes
    some time and doesn't make a huge difference in most cases.

    >>> od = nwa.od_cost_matrix(points, points)
    >>> nwa.rules.split_lines = True
    >>> od = nwa.od_cost_matrix(points, points)

    >>> nwa.log[['split_lines', 'percent_missing', 'cost_mean']]
       split_lines  percent_missing  cost_mean
    0        False           0.9966  15.270462
    1         True           0.8973  15.249900

    >>> nwa.rules.split_lines = False

    Setting a high search_tolerance will make faraway points find their way to the
    network.

    >>> for i in [100, 250, 500, 1000]:
    ...     nwa.rules.search_tolerance = i
    ...     od = nwa.od_cost_matrix(points, points)

    >>> nwa.log.iloc[-4:][['percent_missing', 'cost_mean', 'search_tolerance', 'search_factor']]
       percent_missing  cost_mean  search_tolerance  search_factor
    2           2.3840  15.235559               100              0
    3           0.9966  15.270462               250              0
    4           0.7976  15.273579               500              0
    5           0.5984  15.268614              1000              0

    High search_tolerance won't affect how the points close to the network are
    connected to network nodes. Points trapped behind deadend oneway streets, can find
    their way out with a higher search_factor.

    >>> nwa.rules.search_tolerance = 250
    >>> for i in [0, 10, 35, 100]:
    ...     nwa.rules.search_factor = i
    ...     od = nwa.od_cost_matrix(points, points)

    >>> nwa.log.iloc[-4:][['percent_missing', 'cost_mean', 'search_tolerance', 'search_factor']]
       percent_missing  cost_mean  search_tolerance  search_factor
    6           0.9966  15.270462               250              0
    7           0.5987  15.063283               250             10
    8           0.4991  14.636172               250             35
    9           0.3994  13.680307               250            100

    The remaining 0.4 percent missing are from/to two points, one on an island with no
    brigde and one at the edge of the road network (would require a larger network).
    These two points only find themselves, and thus has 999 missing values.

    >>> n_missing = od.groupby("origin").minutes.agg(lambda x: x.isna().sum())
    >>> n_missing.n_largest(3)
    59     999
    510    999
    0        2
    Name: minutes, dtype: int64

    By default, the distance from origin/destination to the network nodes is given a
    weight of 0. This means, if the search_tolerance is high, points far away from the
    network will get unrealisticly low travel times/distances. The weight from origin/
    destination to the network nodes can be set with the 'nodedist_kmh' parameter if
    the weight is 'minutes', and the 'nodedist_multiplier' if the weight is 'meters'.

    If the weight is 'minutes', setting 'nodedist_kmh' to 5 means a distance of 1000
    meters will get a weight of 12 minutes.

    >>> nwa.rules.search_tolerance = 5000
    >>> for i in [3, 10, 50]:
    ...     nwa.rules.nodedist_kmh = i
    ...     od = nwa.od_cost_matrix(points, points)
    ...
    >>> nwa.log.iloc[-3:][['nodedist_kmh', 'cost_mean']]
       nodedist_kmh  cost_mean
    10            3  15.924197
    11           10  14.817717
    12           50  13.964457

    If the weight is 'meters', setting nodedist_multiplier=1 will make the distance
    to nodes count as its straight line distance.

    >>> rules = NetworkAnalysisRules(
    ...     weight="meters",
    ...     search_tolerance=5000,
    ... )
    >>> nwa = NetworkAnalysis(network=directed_roads, rules=rules)
    >>> od = nwa.od_cost_matrix(points, points)
    >>> nwa.rules.nodedist_multiplier = 1
    >>> od = nwa.od_cost_matrix(points, points)

    >>> nwa.log[['nodedist_multiplier', 'cost_mean']]
      nodedist_multiplier     cost_mean
    0                None  10228.400228
    1                   1  10277.926186
    """

    directed: bool
    weight: str
    search_tolerance: int = 250
    search_factor: int = 0
    split_lines: bool = True
    nodedist_multiplier: int | float | None = None
    nodedist_kmh: int | float | None = None

    def _update_rules(self):
        """Stores the rules as separate attributes.

        Used for checking whether the rules have changed and the graph have to be
        remade.
        """
        self._directed = self.directed
        self._weight = self.weight
        self._search_tolerance = self.search_tolerance
        self._search_factor = self.search_factor
        self._split_lines = self.split_lines
        self._nodedist_multiplier = self.nodedist_multiplier
        self._nodedist_kmh = self.nodedist_kmh

    def _rules_have_changed(self):
        """Checks if any of the rules have changed since the graph was last created.

        If no rules have changed, time can be saved by not remaking the graph
        (the network and the points have to be unchanged as well).
        """
        if self.directed != self._directed:
            return True
        if self.weight != self._weight:
            return True
        if self.search_factor != self._search_factor:
            return True
        if self.search_tolerance != self._search_tolerance:
            return True
        if self.split_lines != self._split_lines:
            return True
        if self.nodedist_multiplier != self._nodedist_multiplier:
            return True
        if self.nodedist_kmh != self._nodedist_kmh:
            return True

    def _validate_weight(self, gdf: GeoDataFrame) -> GeoDataFrame:
        if "meter" in self.weight or "metre" in self.weight and unit_is_meters(gdf):
            if self.nodedist_kmh:
                raise ValueError("Cannot set 'nodedist_kmh' when 'weight' is meters.")
            gdf[self.weight] = gdf.length
            return gdf

        # allow abbreviation of 'minutes' to be nice
        elif (
            self.weight == "min" or "minut" in self.weight and "minutes" in gdf.columns
        ):
            if self.nodedist_multiplier:
                raise ValueError(
                    "Cannot set 'nodedist_multiplier' when 'weight' is minutes. "
                    "Set 'nodedist_kmh' instead."
                )
            self.weight = "minutes"
            gdf["minutes"] = gdf[self.weight]
            self._check_for_nans(gdf, self.weight)
            gdf = self._try_to_float(gdf, self.weight)
            self._check_for_negative_values(gdf, self.weight)
            return gdf

        elif self.weight in gdf.columns:
            self._check_for_nans(gdf, self.weight)
            gdf = self._try_to_float(gdf, self.weight)
            self._check_for_negative_values(gdf, self.weight)
            return gdf

        # at this point, the weight is wrong. Now to determine the error
        # message

        if "meter" in self.weight or "metre" in self.weight:
            raise ValueError(
                "the crs of the roads have to have units in 'meters' when the "
                "weight is 'meters'."
            )

        if self.weight == "minutes":
            incorrect_weight_column = (
                "Cannot find 'weight' column for minutes. "
                "Try running one of the 'make_directed_network_' methods"
                ", or set 'weight' to 'meters'"
            )

        else:
            incorrect_weight_column = f"Cannot find 'weight' column {self.weight}"

        raise KeyError(incorrect_weight_column)

    @staticmethod
    def _check_for_nans(df, col):
        """Remove NaNs and give warning if there are any."""
        if all(df[col].isna()):
            raise ValueError(f"All values in the {col!r} column are NaN.")

        nans = sum(df[col].isna())
        if nans:
            raise ValueError(
                f"{nans} rows have missing values in the {col!r} column. "
                "Fill these rows with 0 or another number.",
            )

    @staticmethod
    def _check_for_negative_values(df, col):
        """Remove negative values and give warning if there are any."""
        negative = sum(df[col] < 0)
        if negative:
            raise ValueError(
                f"{negative} negative values found in the {col!r} column. Fill these "
                "with a number greater than or equal to zero.",
            )

    @staticmethod
    def _try_to_float(df, col):
        """Try to convert weight column to float, raise ValueError if it fails."""
        try:
            df[col] = df[col].astype(float)
        except ValueError as e:
            raise ValueError(
                f"The {col!r} column must be numeric. Got characters that couldn't be "
                "interpreted as numbers."
            ) from e
        return df

    def copy(self):
        return copy(self)

    def deepcopy(self):
        return deepcopy(self)



import numbers
from collections.abc import Iterator, Sized
from typing import Any

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import shapely
from geopandas import GeoDataFrame, GeoSeries
from pandas.api.types import is_array_like, is_dict_like, is_list_like
from shapely import Geometry, box, wkb, wkt
from shapely.geometry import Point
from shapely.ops import unary_union


def to_geoseries(obj: Any, crs: Any | None = None) -> GeoSeries:
    if crs is None:
        try:
            crs = obj.crs
        except AttributeError:
            pass

    try:
        if hasattr(obj.index, "values"):
            # pandas objects
            index = obj.index
        else:
            # list
            index = None
    except AttributeError:
        index = None

    try:
        # this works for geodataframe, geoseries and DataFrame with geometry column
        obj = obj.geometry.values
    except AttributeError:
        try:
            # if pandas series
            obj = obj.values
        except AttributeError:
            # geoseries will raise an Exception for non-geometry objects,
            # so we can safely pass here
            pass

    return GeoSeries(obj, index=index, crs=crs)


def to_shapely(obj) -> Geometry:
    if isinstance(obj, Geometry):
        return obj
    if not hasattr(obj, "__iter__"):
        raise TypeError(type(obj))
    if hasattr(obj, "unary_union"):
        return obj.unary_union
    if is_bbox_like(obj):
        return box(*obj)
    try:
        return Point(*obj)
    except TypeError as e:
        raise TypeError(obj) from e


def from_4326(lon: float, lat: float, crs=25833):
    """Get utm 33 N coordinates from lonlat (4326)."""
    transformer = pyproj.Transformer.from_crs(
        "EPSG:4326", f"EPSG:{crs}", always_xy=True
    )
    return transformer.transform(lon, lat)


def to_4326(lon: float, lat: float, crs=25833):
    """Get degree coordinates  33 N coordinates from lonlat (4326)."""
    transformer = pyproj.Transformer.from_crs(
        f"EPSG:{crs}", "EPSG:4326", always_xy=True
    )
    return transformer.transform(lon, lat)


def coordinate_array(
    gdf: GeoDataFrame | GeoSeries,
    strict=False,
) -> np.ndarray[np.ndarray[float], np.ndarray[float]]:
    """Creates a 2d ndarray of coordinates from point geometries.

    Args:
        gdf: GeoDataFrame or GeoSeries of point geometries.

    Returns:
        np.ndarray of np.ndarrays of coordinates.

    Examples
    --------
    >>> from sgis import coordinate_array, random_points
    >>> points = random_points(5)
    >>> points
                    geometry
    0  POINT (0.59376 0.92577)
    1  POINT (0.34075 0.91650)
    2  POINT (0.74841 0.10627)
    3  POINT (0.00966 0.87868)
    4  POINT (0.38046 0.87879)
    >>> coordinate_array(points)
    array([[0.59376221, 0.92577159],
        [0.34074678, 0.91650446],
        [0.74840912, 0.10626954],
        [0.00965935, 0.87867915],
        [0.38045827, 0.87878816]])
    >>> coordinate_array(points.geometry)
    array([[0.59376221, 0.92577159],
        [0.34074678, 0.91650446],
        [0.74840912, 0.10626954],
        [0.00965935, 0.87867915],
        [0.38045827, 0.87878816]])
    """
    if isinstance(gdf, GeoDataFrame):
        gdf = gdf.geometry
    if strict:
        return np.array([(geom.x, geom.y) for geom in gdf])
    return np.array(
        [(geom.x, geom.y) if hasattr(geom, "x") else (None, None) for geom in gdf]
    )


def to_gdf(
    obj: Geometry
    | str
    | bytes
    | list
    | tuple
    | dict
    | GeoSeries
    | pd.Series
    | pd.DataFrame
    | Iterator,
    crs: str | tuple[str] | None = None,
    geometry: str | tuple[str] | int | None = None,
    **kwargs,
) -> GeoDataFrame:
    """Converts geometry-like objects to a GeoDataFrame.

    Constructs a GeoDataFrame from any geometry-like object (coordinates, wkt, wkb),
    or any interable of such objects.

    Meant for convenience in testing and exploring, not for production code since it
    introduces unnecessary overhead.

    If obj is a DataFrame or dictionary, geometries can be in one column/key or 2-3
    if coordiantes are in x and x (and z) columns. The column/key "geometry" is used
    by default if it exists. The index and other columns/keys are preserved.

    Args:
        obj: the object to be converted to a GeoDataFrame.
        crs: if None (default), it uses the crs of the GeoSeries if GeoSeries
            is the input type. Otherwise, no crs is used.
        geometry: Name of column(s) or key(s) in obj with geometry-like values.
            If not specified, the key/column 'geometry' will be used if it
            exists. If multiple columns, can be given as e.g. "xyz" or ["x", "y"].
        **kwargs: additional keyword arguments taken by the GeoDataFrame constructor.

    Returns:
        A GeoDataFrame with one column, the geometry column.

    Examples
    --------
    >>> from sgis import to_gdf
    >>> coords = (10, 60)
    >>> to_gdf(coords, crs=4326)
                        geometry
    0  POINT (10.00000 60.00000)

    From wkt.

    >>> wkt = "POINT (10 60)"
    >>> to_gdf(wkt, crs=4326)
                        geometry
    0  POINT (10.00000 60.00000)

    From DataFrame with x, y (optionally z) coordinate columns. Index and
    columns are preserved.

    >>> df = pd.DataFrame({"x": [10, 11], "y": [60, 59]}, index=[1,3])
        x   y
    1  10  60
    3  11  59
    >>> gdf = to_gdf(df, geometry=["x", "y"], crs=4326)
    >>> gdf
        x   y                   geometry
    1  10  60  POINT (10.00000 60.00000)
    3  11  59  POINT (11.00000 59.00000)

    For DataFrame/dict with a geometry-like column named "geometry". If the column has
    another name, it must be set with the geometry parameter.

    >>> df = pd.DataFrame({"col": [1, 2], "geometry": ["point (10 60)", (11, 59)]})
    >>> df
       col       geometry
    0    1  point (10 60)
    1    2       (11, 59)
    >>> gdf = to_gdf(df, crs=4326)
    >>> gdf
       col                   geometry
    0    1  POINT (10.00000 60.00000)
    1    2  POINT (11.00000 59.00000)

    From Series.

    >>> series = Series({1: (10, 60), 3: (11, 59)})
    >>> to_gdf(series)
                        geometry
    1  POINT (10.00000 60.00000)
    3  POINT (11.00000 59.00000)

    Multiple coordinates will be converted to points, unless a line or polygon geometry
    is constructed beforehand.

    >>> coordslist = [(10, 60), (11, 59)]
    >>> to_gdf(coordslist, crs=4326)
                        geometry
    0  POINT (10.00000 60.00000)
    1  POINT (11.00000 59.00000)

    >>> from shapely.geometry import LineString
    >>> to_gdf(LineString(coordslist), crs=4326)
                                                geometry
    0  LINESTRING (10.00000 60.00000, 11.00000 59.00000)

    From 2 or 3 dimensional array.

    >>> arr = np.random.randint(100, size=(5, 3))
    >>> to_gdf(arr)
                             geometry
    0  POINT Z (82.000 88.000 82.000)
    1  POINT Z (70.000 92.000 20.000)
    2   POINT Z (91.000 34.000 3.000)
    3   POINT Z (1.000 50.000 77.000)
    4  POINT Z (58.000 49.000 46.000)
    """
    if isinstance(obj, GeoDataFrame):
        if not crs:
            return obj
        if not obj.crs:
            return obj.set_crs(crs)
        return obj.to_crs(crs)

    if obj is None:
        raise TypeError("Cannot convert NoneType to GeoDataFrame.")

    if isinstance(obj, GeoSeries):
        geom_col = geometry or "geometry"
        return _geoseries_to_gdf(obj, geom_col, crs, **kwargs)

    if is_array_like(geometry) and len(geometry) == len(obj):
        geometry = GeoSeries(
            _make_one_shapely_geom(g) for g in geometry if g is not None
        )
        return GeoDataFrame(obj, geometry=geometry, crs=crs, **kwargs)

    geom_col: str = find_geometry_column(obj, geometry)
    index = kwargs.pop("index", None)

    # get done with iterators that get consumed by 'all'
    if isinstance(obj, Iterator) and not isinstance(obj, Sized):
        obj = GeoSeries(
            (_make_one_shapely_geom(g) for g in obj if g is not None), index=index
        )
        return GeoDataFrame({geom_col: obj}, geometry=geom_col, crs=crs, **kwargs)

    if hasattr(obj, "__len__") and not len(obj):
        return GeoDataFrame({"geometry": []}, crs=crs)

    crs = crs or get_crs_from_dict(obj)

    if not is_dict_like(obj):
        if is_bbox_like(obj):
            obj = GeoSeries(shapely.box(*obj), index=index)
            return GeoDataFrame({geom_col: obj}, geometry=geom_col, crs=crs, **kwargs)
        if is_nested_geojson(obj):
            # crs = crs or get_crs_from_dict(obj)
            obj = pd.concat(
                (GeoSeries(_from_json(g)) for g in obj if g is not None),
                ignore_index=True,
            )
            if index is not None:
                obj.index = index
            return GeoDataFrame({geom_col: obj}, geometry=geom_col, crs=crs, **kwargs)
        # list etc.
        else:
            obj = GeoSeries(make_shapely_geoms(obj), index=index)
            return GeoDataFrame(
                {geom_col: obj}, geometry=geom_col, index=index, crs=crs, **kwargs
            )

    # now we have dict, Series or DataFrame

    obj = obj.copy()

    # preserve Series/DataFrame index
    index = obj.index if hasattr(obj, "index") and index is None else index

    if geom_col in obj.keys():
        if isinstance(obj, pd.DataFrame):
            notna = obj[geom_col].notna()
            obj.loc[notna, geom_col] = list(
                make_shapely_geoms(obj.loc[notna, geom_col])
            )
            obj[geom_col] = GeoSeries(obj[geom_col])
            return GeoDataFrame(obj, geometry=geom_col, crs=crs, **kwargs)
        if isinstance(obj[geom_col], Geometry):
            return GeoDataFrame(
                dict(obj), geometry=geom_col, crs=crs, index=[0], **kwargs
            )
        if not hasattr(obj[geom_col], "__iter__") or len(obj[geom_col]) == 1:
            obj[geom_col] = make_shapely_geoms(obj[geom_col])
            return GeoDataFrame(
                dict(obj), geometry=geom_col, crs=crs, index=index, **kwargs
            )
        obj[geom_col] = GeoSeries(make_shapely_geoms(obj[geom_col]), index=index)
        return GeoDataFrame(dict(obj), geometry=geom_col, crs=crs, **kwargs)

    if geometry and all(g in obj for g in geometry):
        obj[geom_col] = _geoseries_from_xyz(obj, geometry, index=index)
        return GeoDataFrame(obj, geometry=geom_col, crs=crs, **kwargs)

    if len(obj.keys()) == 1:
        key = list(obj.keys())[0]
        if isinstance(obj, dict):
            geoseries = GeoSeries(
                make_shapely_geoms(list(obj.values())[0]), index=index
            )
        elif isinstance(obj, pd.Series):
            geoseries = GeoSeries(make_shapely_geoms(obj), index=index)
        else:
            geoseries = GeoSeries(make_shapely_geoms(obj.iloc[:, 0]), index=index)
        return GeoDataFrame({key: geoseries}, geometry=key, crs=crs, **kwargs)

    if geometry and geom_col not in obj or isinstance(obj, pd.DataFrame):
        raise ValueError("Cannot find geometry column(s)", geometry)

    # geojson, __geo_interface__
    if (
        isinstance(obj, dict)
        and sum(key in obj for key in ["type", "coordinates", "features"]) >= 2
    ):
        if "geometry" in obj:
            geometry = "geometry"

        # crs = crs or get_crs_from_dict(obj)
        obj = GeoSeries(_from_json(obj), index=index)
        return GeoDataFrame({geom_col: obj}, geometry=geom_col, crs=crs, **kwargs)

    try:
        geoseries = _series_like_to_geoseries(obj, index=index)
    except ValueError:
        geoseries = _series_like_to_geoseries(obj.dropna(), index=obj.dropna().index)
    return GeoDataFrame(geometry=geoseries, crs=crs, **kwargs)


def make_shapely_geoms(obj):
    if _is_one_geometry(obj):
        return _make_one_shapely_geom(obj)
    if isinstance(obj, dict) and "coordinates" in obj:
        return _from_json(obj)
    return (_make_one_shapely_geom(g) for g in obj)


"""def is_boundingbox(obj) -> bool:
    if not hasattr(obj, "__iter__"):
        return False

    classname = obj.__class__.__name__.lower()
    if "bounding" not in classname and "box" not in classname:
        return False

    if len(obj) == 4 and all(isinstance(x, numbers.Number) for x in obj):
        return True

    return False"""


def is_bbox_like(obj) -> bool:
    if (
        hasattr(obj, "__iter__")
        and len(obj) == 4
        and all(isinstance(x, numbers.Number) for x in obj)
    ):
        return True
    return False


def is_nested_geojson(obj) -> bool:
    if hasattr(obj, "__iter__") and all(isinstance(g, dict) for g in obj):
        return True
    return False


def get_crs_from_dict(obj):
    if (
        not hasattr(obj, "__iter__")
        or not is_dict_like(obj)
        and not is_dict_like(obj[0])
    ):
        return None

    if not is_dict_like(obj) and is_dict_like(obj[0]):
        crss = list({get_crs_from_dict(g) for g in obj})
        return crss[0] if len(crss) == 1 else None

    if "properties" in obj:
        return get_crs_from_dict(obj["properties"])

    if "crs" in obj:
        obj = obj["crs"]
        while is_dict_like(obj):
            if "properties" in obj:
                obj = obj["properties"]
            elif "name" in obj:
                obj = obj["name"]
            else:
                return None
        return obj

    return None


def _from_json(obj: dict):
    if not isinstance(obj, dict) and isinstance(obj[0], dict):
        return [_from_json(g) for g in obj]
    if "geometry" in obj:
        return _from_json(obj["geometry"])
    if "features" in obj:
        return _from_json(obj["features"])
    coords = obj["coordinates"]
    constructor = eval("shapely.geometry." + obj.get("type", Point))
    try:
        return constructor(coords)
    except TypeError:
        while len(coords) == 1:
            coords = coords[0]
        return constructor(coords)


def _series_like_to_geoseries(obj, index):
    if index is None:
        index = obj.keys()
    if isinstance(obj, dict):
        return GeoSeries(make_shapely_geoms(list(obj.values())), index=index)
    else:
        return GeoSeries(make_shapely_geoms(obj.values), index=index)


def _geoseries_to_gdf(obj: GeoSeries, geometry, crs, **kwargs) -> GeoDataFrame:
    if not crs:
        crs = obj.crs
    else:
        obj = obj.to_crs(crs) if obj.crs else obj.set_crs(crs)

    return GeoDataFrame({geometry: obj}, geometry=geometry, crs=crs, **kwargs)


def find_geometry_column(obj, geometry) -> str:
    if geometry is None:
        return "geometry"

    # dict key
    if not is_list_like(geometry) and geometry in obj:
        return geometry

    # nested dict key
    if len(geometry) == 1 and geometry[0] in obj:
        return geometry[0]

    if len(geometry) in {2, 3}:
        return "geometry"

    raise ValueError(
        "geometry should be a geometry column or x, y (z) coordinate columns."
    )


def _geoseries_from_xyz(obj, geometry, index) -> GeoSeries:
    """Make geoseries from the geometry column or columns (x y (z))."""

    if len(geometry) == 2:
        x, y = geometry
        z = None

    elif len(geometry) == 3:
        x, y, z = geometry
        z = obj[z]

    else:
        raise ValueError(
            "geometry should be a geometry column or x, y (z) coordinate columns."
        )

    return gpd.GeoSeries.from_xy(x=obj[x], y=obj[y], z=z, index=index)


def _is_one_geometry(obj) -> bool:
    if (
        isinstance(obj, (str, bytes, Geometry))
        or all(isinstance(i, numbers.Number) for i in obj)
        or not hasattr(obj, "__iter__")
    ):
        return True
    return False


def _make_one_shapely_geom(obj):
    """Create shapely geometry from wkt, wkb or coordinate tuple.

    Works recursively if the object is a nested iterable.
    """
    if isinstance(obj, str):
        return wkt.loads(obj)

    if isinstance(obj, bytes):
        return wkb.loads(obj)

    if isinstance(obj, Geometry):
        return obj

    if not hasattr(obj, "__iter__"):
        raise ValueError(
            f"Couldn't create shapely geometry from {obj} of type {type(obj)}"
        )

    if isinstance(obj, GeoSeries):
        raise TypeError(
            "to_gdf doesn't accept iterable of GeoSeries. Instead use: "
            "pd.concat(to_gdf(obj) for obj in geoseries_iterable)"
        )

    if not any(isinstance(g, numbers.Number) for g in obj):
        # we're likely dealing with a nested iterable, so let's
        # recursively dig down to the coords/wkt/wkb
        if len(obj) == 2 or len(obj) == 3:
            try:
                obj = [float(g) for g in obj]
                return Point(obj)
            except Exception:
                pass
        return unary_union([_make_one_shapely_geom(g) for g in obj])

    elif len(obj) == 2 or len(obj) == 3:
        return Point(obj)

    elif len(obj) == 4:
        return box(*obj)
    else:
        raise ValueError(
            "If 'geom' is an iterable, each item should consist of "
            "wkt, wkb or (x, y (z) or bbox). Got ",
            obj,
        )



from collections.abc import Callable

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import Polygon

from ..helpers import get_non_numpy_func_name, get_numpy_func


def prepare_zonal(polygons: gpd.GeoDataFrame, aggfunc):
    polygons = polygons.reset_index(drop=True)[["geometry"]]

    if isinstance(aggfunc, str) or callable(aggfunc):
        aggfunc = [aggfunc]

    aggfunc = [f if callable(f) else get_numpy_func(f) for f in aggfunc]

    func_names = [get_non_numpy_func_name(f) for f in aggfunc]

    return polygons, aggfunc, func_names


def make_geometry_iterrows(gdf):
    """Because pandas iterrows returns non-geo Series"""
    return list(gdf.geometry.items())


def zonal_func(
    poly_iter: tuple[int, Polygon],
    cube,
    array_func: Callable,
    aggfunc,
    func_names,
    by_date,
) -> pd.DataFrame:
    cube = cube.copy()
    i, polygon = poly_iter
    if not by_date or cube.df["date"].isna().all():
        df = _clip_and_aggregate(
            cube, polygon, array_func, aggfunc, func_names, date=None, i=i
        )
        return df if by_date else df.drop(columns="date")

    out = []

    na_date = cube.query("date.isna()")
    df = _clip_and_aggregate(
        na_date, polygon, array_func, aggfunc, func_names, date=pd.NA, i=i
    )
    out.append(df)

    cube.df = cube.df[lambda x: x["date"].notna()]

    for dt in cube.df["date"].unique():
        cube_date = cube.query(f"date == {dt}")
        df = _clip_and_aggregate(
            cube_date, polygon, array_func, aggfunc, func_names, dt, i
        )
        out.append(df)

    return pd.concat(out)


def _no_overlap_df(func_names, i, date):
    df = pd.DataFrame(columns=func_names, index=[i])
    df["date"] = date
    df["_no_overlap"] = 1
    return df


def _clip_and_aggregate(cube, polygon, array_func, aggfunc, func_names, date, i):
    if not len(cube):
        return _no_overlap_df(func_names, i, date)
    clipped = cube.clipmerge(polygon)
    if not len(clipped) or clipped.arrays.isna().all():
        return _no_overlap_df(func_names, i, date)
    assert len(clipped) == 1
    array = clipped[0].array
    df = _aggregate(array, array_func, aggfunc, func_names, date, i)
    return df


def _aggregate(array, array_func, aggfunc, func_names, date, i):
    if array_func:
        array = array_func(array)
    # flat_array = array.astype(np.float64).flatten()
    flat_array = array.flatten()
    no_nans = flat_array[~np.isnan(flat_array)]
    data = {}
    for f, name in zip(aggfunc, func_names, strict=True):
        num = f(no_nans)
        data[name] = num
    df = pd.DataFrame(data, index=[i])
    df["date"] = date
    df["_no_overlap"] = 0
    return df


def zonal_post(
    aggregated: list[pd.DataFrame], polygons, idx_mapper, idx_name, dropna
) -> pd.DataFrame:
    out = gpd.GeoDataFrame(
        pd.concat(aggregated), geometry=polygons.geometry.values, crs=polygons.crs
    ).sort_index()

    out.index = out.index.map(idx_mapper)
    out.index.name = idx_name

    if dropna:
        out = out.loc[~out.drop(columns="geometry").isna().all(axis=1)]
        return out.loc[out["_no_overlap"] != 1].drop(columns="_no_overlap")
    else:
        return out.drop(columns="_no_overlap")



"""Function to check whether we are in Dapla without importing Dapla.

Function from: From https://github.com/statisticsnorway/ssb-altinn-python/blob/main/src/altinn/utils.py
"""

import os


def is_dapla() -> bool:
    """From https://github.com/statisticsnorway/ssb-altinn-python/blob/main/src/altinn/utils.py."""
    jupyter_image_spec = os.environ.get("JUPYTER_IMAGE_SPEC")
    return bool(jupyter_image_spec and "dapla-jupyterlab" in jupyter_image_spec)



import re
import warnings

import numpy as np
import pandas as pd
import shapely
from geopandas import GeoDataFrame, GeoSeries
from geopandas.array import GeometryArray
from numpy.typing import NDArray
from shapely import (
    extract_unique_points,
    force_2d,
    get_coordinates,
    get_exterior_ring,
    get_parts,
    linearrings,
    linestrings,
    make_valid,
    polygons,
)
from shapely.errors import GEOSException
from shapely.geometry import LinearRing, LineString, Point

from ..networkanalysis.closing_network_holes import get_angle
from .buffer_dissolve_explode import buff, dissexp
from .conversion import coordinate_array, to_geoseries
from .duplicates import get_intersections, update_geometries
from .general import (
    clean_geoms,
    sort_large_first,
    sort_long_first,
    sort_small_first,
    to_lines,
)
from .geometry_types import get_geom_type, make_all_singlepart, to_single_geom_type
from .overlay import clean_overlay
from .polygon_operations import (
    close_all_holes,
    eliminate_by_longest,
    get_cluster_mapper,
    get_gaps,
)
from .polygons_as_rings import PolygonsAsRings
from .sfilter import sfilter, sfilter_inverse


warnings.simplefilter(action="ignore", category=UserWarning)
warnings.simplefilter(action="ignore", category=RuntimeWarning)


PRECISION = 1e-4
BUFFER_RES = 50


def coverage_clean(
    gdf: GeoDataFrame,
    tolerance: int | float,
    duplicate_action: str = "fix",
    # spike_action: str = "ignore",
    grid_sizes: tuple[None | int] = (
        None,
        # 1e-6,
        # 1e-5,
        # 1e-4,
    ),
) -> GeoDataFrame:
    """Fix thin gaps, holes, slivers and double surfaces.

    Rules:
    - Holes (interiors) thinner than the tolerance are closed.
    - Gaps between polygons are filled if thinner than the tolerance.
    - Sliver polygons thinner than the tolerance are eliminated
    into the neighbor polygon with the longest shared border.
    - Double surfaces thinner than the tolerance are eliminated.
    If duplicate_action is "fix", thicker double surfaces will
    be updated from top to bottom of the GeoDataFrame's rows.
    - Line and point geometries are removed.
    - MultiPolygons are exploded to Polygons.
    - Index is reset.

    Args:
        gdf: GeoDataFrame to be cleaned.
        tolerance: distance (usually meters) used as the minimum thickness
            for polygons to be eliminated. Any gap, hole, sliver or double
            surface that are empty after a negative buffer of tolerance / 2
            are eliminated into the neighbor with the longest shared border.
        duplicate action: Either "fix", "error" or "ignore".
            If "fix" (default), double surfaces thicker than the
            tolerance will be updated from top to bottom (function update_geometries)
            and then dissolved into the neighbor polygon with the longest shared border.
            If "error", an Exception is raised if there are any double surfaces thicker
            than the tolerance. If "ignore", double surfaces are kept as is.
        spike_action: Either "fix", "ignore" or "try".

    Returns:
        A GeoDataFrame with cleaned polygons.

    """

    _cleaning_checks(gdf, tolerance, duplicate_action)

    if not gdf.index.is_unique:
        gdf = gdf.reset_index(drop=True)

    gdf = make_all_singlepart(gdf).loc[
        lambda x: x.geom_type.isin(["Polygon", "MultiPolygon"])
    ]

    gdf = clean_geoms(gdf)

    gdf.geometry = shapely.simplify(gdf.geometry, PRECISION)

    gdf = (
        clean_geoms(gdf)
        .pipe(make_all_singlepart)
        .loc[lambda x: x.geom_type.isin(["Polygon", "MultiPolygon"])]
    )

    try:
        gaps = get_gaps(gdf, include_interiors=True)
    except GEOSException:
        for i, grid_size in enumerate(grid_sizes):
            try:
                gaps = get_gaps(gdf, include_interiors=True, grid_size=grid_size)
                break
            except GEOSException as e:
                if i == len(grid_sizes) - 1:
                    explore_geosexception(e, gdf)
                    raise e

    if duplicate_action == "ignore":
        double = GeoDataFrame({"geometry": []}, crs=gdf.crs)
        double["_double_idx"] = None
    else:
        double = get_intersections(gdf)
        double["_double_idx"] = range(len(double))

    gdf, slivers = split_out_slivers(gdf, tolerance)

    thin_gaps_and_double = pd.concat([gaps, double]).loc[
        lambda x: x.buffer(-tolerance / 2).is_empty
    ]

    all_are_thin = double["_double_idx"].isin(thin_gaps_and_double["_double_idx"]).all()

    if not all_are_thin and duplicate_action == "fix":
        gdf, thin_gaps_and_double, slivers = _properly_fix_duplicates(
            gdf, double, slivers, thin_gaps_and_double, tolerance
        )

    elif not all_are_thin and duplicate_action == "error":
        raise ValueError("Large double surfaces.")

    to_eliminate = pd.concat([thin_gaps_and_double, slivers], ignore_index=True)
    to_eliminate.geometry = shapely.simplify(to_eliminate.geometry, PRECISION)

    # eliminate super-thin slivers causing weird geometries
    is_thin = to_eliminate.buffer(-PRECISION).is_empty
    thick, thin = to_eliminate[~is_thin], to_eliminate[is_thin]
    for i, grid_size in enumerate(grid_sizes):
        try:
            to_eliminate = eliminate_by_longest(
                thick,
                thin,
                remove_isolated=False,
                ignore_index=True,
                grid_size=grid_size,
            )
            break
        except GEOSException as e:
            if i == len(grid_sizes) - 1:
                explore_geosexception(e, gdf, thick, thin)
                raise e

    to_eliminate = to_eliminate.loc[lambda x: ~x.buffer(-PRECISION / 10).is_empty]

    to_eliminate["_eliminate_idx"] = range(len(to_eliminate))
    gdf["_poly_idx"] = range(len(gdf))

    to_eliminate["_cluster"] = get_cluster_mapper(to_eliminate.buffer(PRECISION))

    gdf_geoms_idx = gdf[["_poly_idx", "geometry"]]

    joined = to_eliminate.sjoin(gdf_geoms_idx, how="left")
    isolated = joined[lambda x: x["_poly_idx"].isna()]
    intersecting = joined[lambda x: x["_poly_idx"].notna()]

    poly_idx_mapper: pd.Series = (
        clean_overlay(
            intersecting[["_eliminate_idx", "geometry"]],
            buff(gdf_geoms_idx, tolerance, resolution=BUFFER_RES),
            geom_type="polygon",
        )
        .pipe(sort_large_first)
        .drop_duplicates("_eliminate_idx")
        .set_index("_eliminate_idx")["_poly_idx"]
    )
    intersecting["_poly_idx"] = intersecting["_eliminate_idx"].map(poly_idx_mapper)

    for i, grid_size in enumerate(grid_sizes):
        try:
            without_double = update_geometries(
                intersecting, geom_type="polygon", grid_size=grid_size
            ).drop(columns=["_eliminate_idx", "_double_idx", "index_right"])
            break
        except GEOSException as e:
            intersecting.geometry = shapely.simplify(
                intersecting.geometry, PRECISION * (10 * i + 1)
            )
            if i == len(grid_sizes) - 1:
                explore_geosexception(e, gdf, intersecting, isolated)
                raise e

    not_really_isolated = isolated.drop(
        columns=[
            "_double_idx",
            "index_right",
        ]
    ).merge(without_double, on="_cluster", how="inner")

    really_isolated = isolated.loc[
        lambda x: ~x["_eliminate_idx"].isin(not_really_isolated["_eliminate_idx"])
    ]

    really_isolated["_poly_idx"] = (
        really_isolated["_cluster"] + gdf["_poly_idx"].max() + 1
    )

    for i, grid_size in enumerate(grid_sizes):
        try:
            cleaned = (
                dissexp(
                    pd.concat([gdf, without_double, isolated, really_isolated]).drop(
                        columns=[
                            "_cluster",
                            "_eliminate_idx",
                            "index_right",
                            "_double_idx",
                        ],
                        errors="ignore",
                    ),
                    by="_poly_idx",
                    aggfunc="first",
                    dropna=True,
                    grid_size=grid_size,
                )
                .sort_index()
                .reset_index(drop=True)
                # .loc[lambda x: ~x.buffer(-PRECISION / 10).is_empty]
            )
            break
        except GEOSException as e:
            if i == len(grid_sizes) - 1:
                explore_geosexception(e, gdf, without_double, isolated, really_isolated)
                raise e

    cleaned.geometry = shapely.make_valid(shapely.simplify(cleaned.geometry, PRECISION))

    for i, grid_size in enumerate(grid_sizes):
        try:
            cleaned = update_geometries(
                cleaned, geom_type="polygon", grid_size=grid_size
            )
            break
        except GEOSException as e:
            cleaned.geometry = shapely.simplify(
                cleaned.geometry, PRECISION * (10 * i + 1)
            )
            if i == len(grid_sizes) - 1:
                explore_geosexception(
                    e, gdf, cleaned, without_double, isolated, really_isolated
                )
                raise e

    missing_indices: pd.Index = sfilter_inverse(
        gdf.representative_point(), cleaned
    ).index

    missing = clean_overlay(
        gdf.loc[missing_indices].drop(columns="_poly_idx"),
        cleaned,
        how="difference",
        geom_type="polygon",
    )

    cleaned = pd.concat([cleaned, missing], ignore_index=True)
    cleaned.geometry = shapely.make_valid(shapely.simplify(cleaned.geometry, PRECISION))

    return cleaned


def split_spiky_polygons(
    gdf: GeoDataFrame, tolerance: int | float, grid_sizes: tuple[None | int] = (None,)
) -> GeoDataFrame:
    if not len(gdf):
        return gdf

    gdf = to_single_geom_type(make_all_singlepart(gdf), "polygon")

    if not gdf.index.is_unique:
        gdf = gdf.reset_index(drop=True)

    polygons_without_spikes = gdf.buffer(tolerance / 2, join_style=2).buffer(
        -tolerance / 2, join_style=2
    )

    donuts_around_polygons = to_lines(
        polygons_without_spikes.to_frame("geometry")
    ).pipe(buff, 1e-3, copy=False)

    def remove_spikes(df):
        df = df.to_frame("geometry")
        df["_ring_idx"] = range(len(df))
        df = df.reset_index(drop=True)

        points = df.copy()
        points.geometry = extract_unique_points(points.geometry)
        points = points.explode(index_parts=False)
        points["_idx"] = range(len(points))

        not_spikes = points.sjoin(donuts_around_polygons).loc[
            lambda x: x["_ring_idx"] == x["index_right"]
        ]
        can_be_polygons = not_spikes.iloc[
            (not_spikes.groupby("_ring_idx").transform("size") >= 3).values
        ]

        without_spikes = (
            can_be_polygons.sort_values("_idx")
            .groupby("_ring_idx")["geometry"]
            .agg(LinearRing)
        )

        missing = df[~df["_ring_idx"].isin(without_spikes.index)].geometry

        return pd.concat([without_spikes, missing]).sort_index()

    without_spikes = GeoDataFrame(
        {
            "geometry": PolygonsAsRings(gdf.geometry)
            .apply_geoseries_func(remove_spikes)
            .to_numpy()
        },
        crs=gdf.crs,
    ).pipe(to_single_geom_type, "polygon")

    is_thin = without_spikes.buffer(-tolerance / 2).is_empty
    without_spikes = pd.concat(
        [
            split_by_neighbors(
                without_spikes[is_thin], without_spikes, tolerance=tolerance
            ),
            without_spikes[~is_thin],
        ]
    )

    for _ in range(2):
        for i, grid_size in enumerate(grid_sizes):
            try:
                without_spikes = update_geometries(
                    sort_small_first(without_spikes), geom_type="polygon"
                )
                break
            except GEOSException as e:
                if i == len(grid_sizes) - 1:
                    raise e

    for i, grid_size in enumerate(grid_sizes):
        try:
            return clean_overlay(
                gdf, without_spikes, how="identity", grid_size=grid_size
            )
        except GEOSException as e:
            if i == len(grid_sizes) - 1:
                raise e


def remove_spikes(gdf: GeoDataFrame, tolerance: int | float) -> GeoDataFrame:
    """Remove thin spikes in polygons.

    Note that this function might be slow. Should only be used if nessecary.

    Args:
        gdf: GeoDataFrame of polygons
        tolerance: distance (usually meters) used as the minimum thickness
            for polygons to be eliminated. Any spike thinner than the tolerance
            will be removed.

    Returns:
        A GeoDataFrame of polygons without spikes thinner.
    """

    gdf.geometry = (
        PolygonsAsRings(gdf.geometry)
        .apply_numpy_func(_remove_spikes, args=(tolerance,))
        .to_numpy()
    )
    return gdf


def _remove_spikes(
    geoms: NDArray[LinearRing], tolerance: int | float
) -> NDArray[LinearRing]:
    if not len(geoms):
        return geoms
    geoms = to_geoseries(geoms).reset_index(drop=True)

    points = (
        extract_unique_points(geoms).explode(index_parts=False).to_frame("geometry")
    )

    points = get_angle_between_indexed_points(points)

    def to_buffered_rings_without_spikes(x):
        polys = GeoSeries(make_valid(polygons(get_exterior_ring(x))))

        return (
            polys.buffer(-tolerance, resolution=BUFFER_RES)
            .explode(index_parts=False)
            .pipe(close_all_holes)
            .pipe(get_exterior_ring)
            .buffer(tolerance * 10)
        )

    buffered = to_buffered_rings_without_spikes(
        geoms.buffer(tolerance / 2, resolution=BUFFER_RES)
    )

    points_without_spikes = (
        extract_unique_points(geoms)
        .explode(index_parts=False)
        .loc[lambda x: x.index.isin(sfilter(x, buffered).index)]
    )

    # linearrings require at least 4 coordinate pairs, or three unique
    points_without_spikes = points_without_spikes.loc[
        lambda x: x.groupby(level=0).size() >= 3
    ]

    # need an index from 0 to n-1 in 'linearrings'
    to_int_index = {
        ring_idx: i
        for i, ring_idx in enumerate(sorted(set(points_without_spikes.index)))
    }
    int_indices = points_without_spikes.index.map(to_int_index)

    as_lines = pd.Series(
        linearrings(
            get_coordinates(points_without_spikes.geometry.values),
            indices=int_indices,
        ),
        index=points_without_spikes.index.unique(),
    )

    # the missing polygons are thin and/or spiky. Let's remove them
    missing = geoms.loc[~geoms.index.isin(as_lines.index)]

    missing = pd.Series(
        [None] * len(missing),
        index=missing.index.values,
    )

    return pd.concat([as_lines, missing]).sort_index()


def get_angle_between_indexed_points(point_df: GeoDataFrame):
    """ "Get angle difference between the two lines"""

    point_df["next"] = point_df.groupby(level=0)["geometry"].shift(-1)

    notna = point_df["next"].notna()

    this = coordinate_array(point_df.loc[notna, "geometry"].values)
    next_ = coordinate_array(point_df.loc[notna, "next"].values)

    point_df.loc[notna, "angle"] = get_angle(this, next_)
    point_df["prev_angle"] = point_df.groupby(level=0)["angle"].shift(1)

    point_df["angle_diff"] = point_df["angle"] - point_df["prev_angle"]

    return point_df


def _properly_fix_duplicates(gdf, double, slivers, thin_gaps_and_double, tolerance):
    # gdf = update_geometries(gdf)
    # gdf, more_slivers = split_out_slivers(gdf, tolerance)
    # slivers = pd.concat([slivers, more_slivers], ignore_index=True)
    # gaps = get_gaps(gdf, include_interiors=True)
    # double = get_intersections(gdf).pipe(update_geometries, geom_type="polygon")
    # double["_double_idx"] = range(len(double))
    # thin_gaps_and_double = pd.concat([gaps, double]).loc[
    #     lambda x: x.buffer(-tolerance / 2).is_empty
    # ]
    # return gdf, thin_gaps_and_double, slivers

    gdf = _dissolve_thick_double_and_update(gdf, double, thin_gaps_and_double)
    gdf, more_slivers = split_out_slivers(gdf, tolerance)
    slivers = pd.concat([slivers, more_slivers], ignore_index=True)
    gaps = get_gaps(gdf, include_interiors=True)
    assert "_double_idx" not in gaps
    double = get_intersections(gdf)
    double["_double_idx"] = range(len(double))
    thin_gaps_and_double = pd.concat([gaps, double], ignore_index=True).loc[
        lambda x: x.buffer(-tolerance / 2).is_empty
    ]

    return gdf, thin_gaps_and_double, slivers


def _dissolve_thick_double_and_update(gdf, double, thin_double):
    large = (
        double.loc[~double["_double_idx"].isin(thin_double["_double_idx"])]
        .drop(columns="_double_idx")
        .pipe(sort_large_first)
        .pipe(update_geometries, geom_type="polygon")
    )
    return (
        clean_overlay(gdf, large, how="update")
        .pipe(sort_large_first)
        .pipe(update_geometries, geom_type="polygon")
    )


def _cleaning_checks(gdf, tolerance, duplicate_action):  # , spike_action):
    if not len(gdf) or not tolerance:
        return gdf
    if tolerance < PRECISION:
        raise ValueError(
            f"'tolerance' must be larger than {PRECISION} to avoid "
            "problems with floating point precision."
        )
    if duplicate_action not in ["fix", "error", "ignore"]:
        raise ValueError("duplicate_action must be 'fix', 'error' or 'ignore'")


def split_out_slivers(
    gdf: GeoDataFrame | GeoSeries, tolerance: float | int
) -> tuple[GeoDataFrame, GeoDataFrame] | tuple[GeoSeries, GeoSeries]:
    is_sliver = gdf.buffer(-tolerance / 2).is_empty
    slivers = gdf.loc[is_sliver]
    gdf = gdf.loc[~is_sliver]
    return gdf, slivers


def split_by_neighbors(df, split_by, tolerance):
    if not len(df):
        return df

    split_by = split_by.copy()
    split_by.geometry = shapely.simplify(split_by.geometry, tolerance)

    intersecting_lines = (
        clean_overlay(to_lines(split_by), buff(df, tolerance), how="identity")
        .pipe(get_line_segments)
        .reset_index(drop=True)
    )

    endpoints = intersecting_lines.boundary.explode(index_parts=False)

    extended_lines = GeoDataFrame(
        {
            "geometry": extend_lines(
                endpoints.loc[lambda x: ~x.index.duplicated(keep="first")].values,
                endpoints.loc[lambda x: ~x.index.duplicated(keep="last")].values,
                distance=tolerance * 3,
            )
        },
        crs=df.crs,
    )

    buffered = buff(extended_lines, tolerance, single_sided=True)

    return clean_overlay(df, buffered, how="identity")


def extend_lines(arr1, arr2, distance):
    if len(arr1) != len(arr2):
        raise ValueError
    if not len(arr1):
        return arr1

    arr1, arr2 = arr2, arr1  # TODO fix

    coords1 = coordinate_array(arr1)
    coords2 = coordinate_array(arr2)

    dx = coords2[:, 0] - coords1[:, 0]
    dy = coords2[:, 1] - coords1[:, 1]
    len_xy = np.sqrt((dx**2.0) + (dy**2.0))
    x = coords1[:, 0] + (coords1[:, 0] - coords2[:, 0]) / len_xy * distance
    y = coords1[:, 1] + (coords1[:, 1] - coords2[:, 1]) / len_xy * distance

    new_points = np.array([None for _ in range(len(arr1))])
    new_points[~np.isnan(x)] = shapely.points(x[~np.isnan(x)], y[~np.isnan(x)])

    new_points[~np.isnan(x)] = make_lines_between_points(
        arr2[~np.isnan(x)], new_points[~np.isnan(x)]
    )
    return new_points


def make_lines_between_points(
    arr1: NDArray[Point], arr2: NDArray[Point]
) -> NDArray[LineString]:
    if arr1.shape != arr2.shape:
        raise ValueError(
            f"Arrays must have equal shape. Got {arr1.shape} and {arr2.shape}"
        )
    coords: pd.DataFrame = pd.concat(
        [
            pd.DataFrame(get_coordinates(arr1), columns=["x", "y"]),
            pd.DataFrame(get_coordinates(arr2), columns=["x", "y"]),
        ]
    ).sort_index()

    return linestrings(coords.values, indices=coords.index)


def get_line_segments(lines) -> GeoDataFrame:
    assert lines.index.is_unique
    if isinstance(lines, GeoDataFrame):
        multipoints = lines.assign(
            **{
                lines._geometry_column_name: force_2d(
                    extract_unique_points(lines.geometry.values)
                )
            }
        )
        return multipoints_to_line_segments(multipoints.geometry)

    multipoints = GeoSeries(extract_unique_points(lines.values), index=lines.index)

    return multipoints_to_line_segments(multipoints)


def multipoints_to_line_segments(multipoints: GeoSeries) -> GeoDataFrame:
    if not len(multipoints):
        return GeoDataFrame({"geometry": multipoints}, index=multipoints.index)

    try:
        crs = multipoints.crs
    except AttributeError:
        crs = None

    try:
        point_df = multipoints.explode(index_parts=False)
        if isinstance(point_df, GeoSeries):
            point_df = point_df.to_frame("geometry")
    except AttributeError:
        points, indices = get_parts(multipoints, return_index=True)
        if isinstance(multipoints.index, pd.MultiIndex):
            indices = pd.MultiIndex.from_arrays(indices, names=multipoints.index.names)

        point_df = pd.DataFrame({"geometry": GeometryArray(points)}, index=indices)

    point_df["next"] = point_df.groupby(level=0)["geometry"].shift(-1)

    first_points = point_df.loc[lambda x: ~x.index.duplicated(), "geometry"]
    is_last_point = point_df["next"].isna()

    point_df.loc[is_last_point, "next"] = first_points
    assert point_df["next"].notna().all()

    point_df["geometry"] = [
        LineString([x1, x2]) for x1, x2 in zip(point_df["geometry"], point_df["next"])
    ]
    return GeoDataFrame(point_df.drop(columns=["next"]), geometry="geometry", crs=crs)


def explore_geosexception(e: GEOSException, *gdfs):
    from ..maps.maps import explore
    from .conversion import to_gdf

    pattern = r"(\d+\.\d+)\s+(\d+\.\d+)"

    matches = re.findall(pattern, str(e))
    coords_in_error_message = [(float(match[0]), float(match[1])) for match in matches]
    exception_point = to_gdf(coords_in_error_message, crs=gdfs[0].crs)
    if len(exception_point):
        exception_point["wkt"] = exception_point.to_wkt()
        explore(exception_point, *gdfs, mask=exception_point.buffer(100))
    else:
        explore(*gdfs)



from collections.abc import Callable
from pathlib import Path

import pandas as pd
from dapla import read_pandas, write_pandas
from geopandas import GeoDataFrame
from pandas import DataFrame

from ..geopandas_tools.general import clean_clip, clean_geoms
from ..geopandas_tools.neighbors import get_neighbor_indices
from .dapla_functions import read_geopandas, write_geopandas


def write_municipality_data(
    data: str | GeoDataFrame | DataFrame,
    out_folder: str,
    municipalities: GeoDataFrame,
    with_neighbors: bool = False,
    muni_number_col: str = "KOMMUNENR",
    file_type: str = "parquet",
    func: Callable | None = None,
    write_empty: bool = False,
    clip: bool = True,
) -> None:
    write_func = (
        _write_neighbor_municipality_data
        if with_neighbors
        else _write_municipality_data
    )

    return write_func(
        data=data,
        out_folder=out_folder,
        municipalities=municipalities,
        muni_number_col=muni_number_col,
        file_type=file_type,
        func=func,
        write_empty=write_empty,
        clip=clip,
    )


def _validate_data(data: str | list[str]) -> str:
    if isinstance(data, (str, Path)):
        return data
    if hasattr(data, "__iter__") and len(data) == 1:
        return data[0]
    elif not isinstance(data, GeoDataFrame):
        raise TypeError("'data' Must be a file path or a GeoDataFrame. Got", type(data))


def _get_out_path(out_folder, muni, file_type):
    return str(Path(out_folder) / f"{muni}.{file_type.strip('.')}")


def _write_municipality_data(
    data: str | GeoDataFrame | DataFrame,
    out_folder: str,
    municipalities: GeoDataFrame,
    muni_number_col: str = "KOMMUNENR",
    file_type: str = "parquet",
    func: Callable | None = None,
    write_empty: bool = False,
    clip: bool = True,
) -> None:
    data = _validate_data(data)

    if isinstance(data, (str, Path)):
        try:
            gdf = read_geopandas(str(data))
        except ValueError as e:
            try:
                gdf = read_pandas(str(data))
            except ValueError:
                raise e.__class__(e, data)
    elif isinstance(data, DataFrame):
        gdf = data
    else:
        raise TypeError(type(data))

    if func is not None:
        gdf = func(gdf)

    gdf = _fix_missing_muni_numbers(gdf, municipalities, muni_number_col, clip)

    for muni in municipalities[muni_number_col]:
        out = _get_out_path(out_folder, muni, file_type)

        gdf_muni = gdf.loc[gdf[muni_number_col] == muni]

        if not len(gdf_muni):
            if write_empty:
                gdf_muni = gdf_muni.drop(columns="geometry", errors="ignore")
                gdf_muni["geometry"] = None
                write_pandas(gdf_muni, out)
            continue

        if not len(gdf_muni):
            if write_empty:
                gdf_muni = gdf_muni.drop(columns="geometry", errors="ignore")
                gdf_muni["geometry"] = None
                write_pandas(gdf_muni, out)
            continue

        write_geopandas(gdf_muni, out)


def _write_neighbor_municipality_data(
    data: str | GeoDataFrame | DataFrame,
    out_folder: str,
    municipalities: GeoDataFrame,
    muni_number_col: str = "KOMMUNENR",
    file_type: str = "parquet",
    func: Callable | None = None,
    write_empty: bool = False,
    clip: bool = True,
) -> None:
    data = _validate_data(data)

    if isinstance(data, (str, Path)):
        gdf = read_geopandas(str(data))

    if func is not None:
        gdf = func(gdf)

    gdf = _fix_missing_muni_numbers(gdf, municipalities, muni_number_col, clip)

    if municipalities.index.name != muni_number_col:
        municipalities = municipalities.set_index(muni_number_col)

    neighbor_munis = get_neighbor_indices(
        municipalities, municipalities, max_distance=1
    )

    for muni in municipalities.index:
        out = _get_out_path(out_folder, muni, file_type)

        muni_and_neighbors = neighbor_munis.loc[[muni]]
        gdf_neighbor = gdf.loc[gdf[muni_number_col].isin(muni_and_neighbors)]

        if not len(gdf_neighbor):
            if write_empty:
                gdf_neighbor["geometry"] = gdf_neighbor["geometry"].astype(str)
                write_pandas(gdf_neighbor, out)
            continue

        write_geopandas(gdf_neighbor, out)


def _fix_missing_muni_numbers(gdf, municipalities, muni_number_col, clip):
    if muni_number_col in gdf and gdf[muni_number_col].notna().all():
        return gdf

    if municipalities is None:
        if muni_number_col not in gdf:
            raise ValueError(
                f"Cannot find column {muni_number_col}. "
                "Specify another column or a municipality GeoDataFrame to clip "
                "the geometries by."
            )
        assert gdf[muni_number_col].isna().any()
        raise ValueError(
            f"Column {muni_number_col} has missing values. Make sure gdf has "
            "correct municipality number info or specify a municipality "
            "GeoDataFrame to clip the geometries by."
        )

    def _clean_overlay(df1, df2):
        return df1.pipe(clean_geoms).overlay(df2, how="intersection").pipe(clean_geoms)

    def _clean_clip(df1, df2, muni_number_col):
        """Looping clip for large datasets because it's faster and safer."""
        all_clipped = []
        for muni in df2[muni_number_col]:
            clipped = clean_clip(df1, df2[df2[muni_number_col] == muni])
            clipped[muni_number_col] = muni
            all_clipped.append(clipped)
        return pd.concat(all_clipped)

    municipalities = municipalities[[muni_number_col, "geometry"]].to_crs(gdf.crs)

    if muni_number_col in gdf and gdf[muni_number_col].isna().any():
        notna = gdf[gdf[muni_number_col].notna()]

        isna = gdf[gdf[muni_number_col].isna()].drop(muni_number_col, axis=1)

        if not clip:
            notna_anymore = isna.sjoin(municipalities).drop(columns="index_right")
        elif len(isna) < 10_000:
            notna_anymore = _clean_overlay(isna, municipalities)
        else:
            notna_anymore = _clean_clip(isna, municipalities, muni_number_col)

        return pd.concat([notna, notna_anymore])

    if not clip:
        return gdf.sjoin(municipalities).drop(columns="index_right")
    if len(gdf) < 10_000:
        return _clean_overlay(gdf, municipalities)
    else:
        return _clean_clip(gdf, municipalities, muni_number_col)
